# -*- coding: utf-8 -*-
# File: _test.py


import sys
import numpy as np
import cv2
import unittest

from .base import ImageAugmentor, AugmentorList
from .imgproc import Contrast
from .noise import SaltPepperNoise
from .misc import Flip, Resize


def _rand_image(shape=(20, 20)):
    return np.random.rand(*shape).astype("float32")


class LegacyBrightness(ImageAugmentor):
    def __init__(self, delta, clip=True):
        super(LegacyBrightness, self).__init__()
        assert delta > 0
        self._init(locals())

    def _get_augment_params(self, _):
        v = self._rand_range(-self.delta, self.delta)
        return v

    def _augment(self, img, v):
        old_dtype = img.dtype
        img = img.astype('float32')
        img += v
        if self.clip or old_dtype == np.uint8:
            img = np.clip(img, 0, 255)
        return img.astype(old_dtype)


class LegacyFlip(ImageAugmentor):
    def __init__(self, horiz=False, vert=False, prob=0.5):
        super(LegacyFlip, self).__init__()
        if horiz and vert:
            raise ValueError("Cannot do both horiz and vert. Please use two Flip instead.")
        elif horiz:
            self.code = 1
        elif vert:
            self.code = 0
        else:
            raise ValueError("At least one of horiz or vert has to be True!")
        self._init(locals())

    def _get_augment_params(self, img):
        h, w = img.shape[:2]
        do = self._rand_range() < self.prob
        return (do, h, w)

    def _augment(self, img, param):
        do, _, _ = param
        if do:
            ret = cv2.flip(img, self.code)
            if img.ndim == 3 and ret.ndim == 2:
                ret = ret[:, :, np.newaxis]
        else:
            ret = img
        return ret

    def _augment_coords(self, coords, param):
        do, h, w = param
        if do:
            if self.code == 0:
                coords[:, 1] = h - coords[:, 1]
            elif self.code == 1:
                coords[:, 0] = w - coords[:, 0]
        return coords


class ImgAugTest(unittest.TestCase):
    def _get_augs(self):
        return AugmentorList([
            Contrast((0.8, 1.2)),
            Flip(horiz=True),
            Resize((30, 30)),
            SaltPepperNoise()
        ])

    def _get_augs_with_legacy(self):
        return AugmentorList([
            LegacyBrightness(0.5),
            LegacyFlip(horiz=True),
            Resize((30, 30)),
            SaltPepperNoise()
        ])

    def test_augmentors(self):
        augmentors = self._get_augs()

        img = _rand_image()
        orig = img.copy()
        tfms = augmentors.get_transform(img)

        # test printing
        print(augmentors)
        print(tfms)

        newimg = tfms.apply_image(img)
        print(tfms)  # lazy ones will instantiate after the first apply

        newimg2 = tfms.apply_image(orig)
        self.assertTrue(np.allclose(newimg, newimg2))
        self.assertEqual(newimg2.shape[0], 30)

        coords = np.asarray([[0, 0], [10, 12]], dtype="float32")
        tfms.apply_coords(coords)

    def test_legacy_usage(self):
        augmentors = self._get_augs()

        img = _rand_image()
        orig = img.copy()
        newimg, tfms = augmentors.augment_return_params(img)
        newimg2 = augmentors.augment_with_params(orig, tfms)
        self.assertTrue(np.allclose(newimg, newimg2))
        self.assertEqual(newimg2.shape[0], 30)

        coords = np.asarray([[0, 0], [10, 12]], dtype="float32")
        augmentors.augment_coords(coords, tfms)

    def test_legacy_augs_new_usage(self):
        augmentors = self._get_augs_with_legacy()

        img = _rand_image()
        orig = img.copy()
        tfms = augmentors.get_transform(img)
        newimg = tfms.apply_image(img)
        newimg2 = tfms.apply_image(orig)
        self.assertTrue(np.allclose(newimg, newimg2))
        self.assertEqual(newimg2.shape[0], 30)

        coords = np.asarray([[0, 0], [10, 12]], dtype="float32")
        tfms.apply_coords(coords)

    def test_legacy_augs_legacy_usage(self):
        augmentors = self._get_augs_with_legacy()

        img = _rand_image()
        orig = img.copy()
        newimg, tfms = augmentors.augment_return_params(img)
        newimg2 = augmentors.augment_with_params(orig, tfms)
        self.assertTrue(np.allclose(newimg, newimg2))
        self.assertEqual(newimg2.shape[0], 30)

        coords = np.asarray([[0, 0], [10, 12]], dtype="float32")
        augmentors.augment_coords(coords, tfms)


if __name__ == '__main__':
    anchors = [(0.2, 0.2), (0.7, 0.2), (0.8, 0.8), (0.5, 0.5), (0.2, 0.5)]
    augmentors = AugmentorList([
        Contrast((0.8, 1.2)),
        Flip(horiz=True),
        # RandomCropRandomShape(0.3),
        SaltPepperNoise()
    ])

    img = cv2.imread(sys.argv[1])
    newimg, prms = augmentors._augment_return_params(img)
    cv2.imshow(" ", newimg.astype('uint8'))
    cv2.waitKey()

    newimg = augmentors._augment(img, prms)
    cv2.imshow(" ", newimg.astype('uint8'))
    cv2.waitKey()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tqdm import tqdm
from loguru import logger

from ..base_model import TransformerModel


class EntityExtractionModel(TransformerModel):
    def __init__(self,
                 args,
                 entity_labels,
                 tokenizer,
                 tagging_type="pointer_sequence"):
        """
        实体抽取模型
        按不同的标注方式构建实体抽取模型

        :param entity_labels:List(str) 实体类型标签列表
        :param tokenizer:tokenizers.Tokenizer
        :param tagging_type:str ['pointer_sequence', 'multi_pointer_sequences', 'multi_heads_matrix', 'span_array', 'crf']

        """
        assert sorted(entity_labels) == sorted(list(set(entity_labels)))

        #  entity_labels = ['[unused1]'] + entity_labels
        #  id2label = {i: label for i, label in enumerate(entity_labels)}
        #  label2id = {label: i for i, label in enumerate(entity_labels)}

        from .pointer_sequence import (PointerSequenceTagger,
                                       PointerSequenceExtractor,
                                       PointerSequenceTrainer)
        #  from ..taggers import PointerSequenceTagger  #, MultiPointerSequencesTagger, MultiHeadsMatrixTagger, SpanArrayTagger, CrfTagger
        #  from ..extractors import PointerSequenceExtractor  #, MultiPointerSequencesExtractor, MultiHeadsMatrixExtractor, SpanArrayExtractor, CrfExtractor
        #  from ..trainers import PointerSequenceTrainer  #, MultiPointerSequencesTraiiner, MultiHeadsMatrixTrainer, SpanArrayTrainer, CrfTrainer
        ner_classes = {
            'pointer_sequence': [
                PointerSequenceTagger, PointerSequenceTrainer,
                PointerSequenceExtractor
            ],
            #  'multi_pointer_sequence': [
            #      MultiPointerSequenceTagger, MultiPointerSequenceTrainer,
            #      MultiPointerSequenceCollector
            #  ],
            #  'multi_heads_matrix': [
            #      MultiHeadsMatrixTagger, MultiHeadsMatrixTrainer,
            #      MultiHeadsMatrixCollector
            #  ],
            #  'span_array':
            #  [SpanArrayTagger, SpanArrayTrainer, SpanArrayCollector],
            #  'crf': [CrfTagger, CrfTrainer, CrfCollector]
        }

        if tagging_type in ner_classes:
            tagger_cls, trainer_cls, extractor_cls = ner_classes[tagging_type]
        else:
            raise ValueError(
                f"tagging_type muse be in ['pointer_sequence', 'multi_pointer_sequences', 'multi_heads_matrix', 'span_array', 'crf']"
            )

        trainer = trainer_cls(args, entity_labels)
        tagger = tagger_cls(tokenizer=tokenizer, label2id=args.label2id)
        extractor = extractor_cls()

        super(EntityExtractionModel, self).__init__(args,
                                                    tagger=tagger,
                                                    trainer=trainer,
                                                    extractor=extractor)
        self.entity_labels = entity_labels
        self.tokenizer = tokenizer
        #  self.label2id = label2id

    def train(self, train_dataflow, eval_dataflow):
        args = self.args

        train_features = self.tagger.encode(train_dataflow,
                                            args.train_max_seq_length)
        #  self.tokenizer,
        #  self.label2id
        #  )
        eval_features = self.tagger.encode(eval_dataflow,
                                           args.eval_max_seq_length)
        #  self.tokenizer, self.label2id)
        self.trainer.train(args, train_features, eval_features)

    def evaluate(self, eval_dataflow, model_path=None):
        args = self.args
        model = self.load_model(model_path=args.best_model_path
                                if model_path is None else model_path)
        eval_features = self.tagger.encode(eval_dataflow,
                                           args.train_max_seq_length)
        self.trainer.evaluate(args, model, eval_features)

    def predict(self, test_dataflow, model_path=None):
        args = self.args

        model = self.load_model(model_path=args.best_model_path
                                if model_path is None else model_path)
        test_features = self.tagger.encode(test_dataflow,
                                           args.train_max_seq_length)
        self.trainer.predict(args, model, test_features)

        from ..utils.ner_utils import save_ner_preds
        save_ner_preds(self.args, self.trainer.preds, test_dataflow)



#  -*- coding: utf-8 -*-
#  File: __init__.py

# https://github.com/celery/kombu/blob/7d13f9b95d0b50c94393b962e6def928511bfda6/kombu/__init__.py#L34-L36
STATICA_HACK = True
globals()['kcah_acitats'[::-1].upper()] = False
if STATICA_HACK:
    from .bsds500 import *
    from .cifar import *
    from .ilsvrc import *
    from .mnist import *
    from .svhn import *
    from .caltech101 import *

from pkgutil import iter_modules
import os
import os.path

__all__ = []


def global_import(name):
    p = __import__(name, globals(), locals(), level=1)
    lst = p.__all__ if '__all__' in dir(p) else dir(p)
    if lst:
        del globals()[name]
        for k in lst:
            if not k.startswith('__'):
                globals()[k] = p.__dict__[k]
                __all__.append(k)


_CURR_DIR = os.path.dirname(__file__)
for _, module_name, _ in iter_modules(
        [_CURR_DIR]):
    srcpath = os.path.join(_CURR_DIR, module_name + '.py')
    if not os.path.isfile(srcpath):
        continue
    if not module_name.startswith('_'):
        global_import(module_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from loguru import logger
from tqdm import tqdm

#  from rich import print


def do_new(args):
    logger.warning(f"do_new")
    latest_dir = args.latest_dir

    import shutil
    if os.path.exists(latest_dir):
        shutil.rmtree(latest_dir)
    os.makedirs(latest_dir)

    import uuid
    local_id = str(uuid.uuid1()).replace('-', '')[:8]
    local_id_file = os.path.join(latest_dir, "local_id")
    with open(local_id_file, 'w') as wt:
        wt.write(f"{local_id}")
    logger.warning(f"New model id: {local_id}")

    args.local_id = local_id
    args.local_dir = os.path.join(args.saved_models_path, args.local_id)
    from .utils import save_args
    save_args(args, latest_dir)

    #  cmd_cp_params =
    #      f"cp {theta_src_path}/templates/{app_type}/{app_type}_params.py {latest_dir}/{dataset_name}_params.py"
    #  logger.warning(f"{cmd_cp_params}")

    logger.warning(f"dataset_name: {args.dataset_name}")
    logger.warning(f"experiment_name: {args.experiment_name}")
    logger.warning(f"local_id: {args.local_id}")
    logger.warning(f"local_dir: {args.local_dir}")
    logger.warning(f"latest_dir: {args.latest_dir}")
    logger.warning(f"saved_models_path: {args.saved_models_path}")


class GlueApp:
    def __init__(self, experiment_params, glue_labels, add_special_args=None):
        self.experiment_params = experiment_params
        self.glue_labels = glue_labels

        from .glue import load_model, get_args

        args = get_args(experiment_params=experiment_params,
                        special_args=[add_special_args])
        args.app_type = 'glue'
        self.args = args
        logger.info(f"args: {args}")
        #  print("[bold cyan]args:[/bold cyan]", args)

        self.trainer = None

    def get_trainer(self):
        # -------------------- Model --------------------
        if self.trainer is None:
            args = self.args

            from .glue import GlueTrainer

            class AppTrainer(GlueTrainer):
                def __init__(self, args, glue_labels):
                    super(AppTrainer, self).__init__(args,
                                                     glue_labels,
                                                     build_model=None)

                #  def on_predict_end(self, args, test_dataset):
                #      super(Trainer, self).on_predict_end(args, test_dataset)

            self.trainer = AppTrainer(args, self.glue_labels)

        return self.trainer

    def load_model(self):
        from .glue import load_model as global_load_model

        model = global_load_model(self.args)
        return model

    def run(self,
            train_data_generator,
            test_data_generator,
            generate_submission=None,
            eval_data_generator=None):
        args = self.args

        assert train_data_generator is not None
        assert test_data_generator is not None

        #  if eval_data_generator is None:
        #      eval_data_generator = train_data_generator

        def do_eda(args):
            from .glue_utils import show_glue_datainfo
            show_glue_datainfo(self.glue_labels, train_data_generator,
                               args.train_file, test_data_generator,
                               args.test_file)

        def do_submit(args):
            submission_file = None
            if generate_submission:
                submission_file = generate_submission(args)
            from .utils import archive_local_model
            archive_local_model(args, submission_file)

        def do_generate_submission(args):
            if generate_submission is not None:
                if args.dataset_file is None:
                    logger.error(f"dataset_file parameter must be set.")
                    return
                if args.submission_file is None:
                    logger.error(f"submission_file parameter must be set.")
                    return
                generate_submission(args,
                                    reviews_file=args.dataset_file,
                                    submission_file=args.submission_file)

        if args.do_new:
            do_new(args)

        elif args.do_eda:
            do_eda(args)

        elif args.do_submit:
            do_submit(args)

        elif args.generate_submission:
            do_generate_submission(args)

        else:

            trainer = self.get_trainer()

            from .glue_utils import load_train_val_examples, load_test_examples

            def do_train(args):
                if eval_data_generator is None:
                    train_examples, val_examples = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.glue_labels,
                        shuffle=True,
                        train_rate=args.train_rate,
                        num_augments=args.num_augments)
                else:
                    train_examples, _ = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.glue_labels,
                        shuffle=True,
                        train_rate=1.0,
                        num_augments=args.num_augments)

                    _, val_examples = load_train_val_examples(
                        args,
                        eval_data_generator,
                        self.glue_labels,
                        shuffle=False,
                        train_rate=0.0,
                        num_augments=0)
                trainer.train(args, train_examples, val_examples)

            def do_eval(args):
                args.model_path = args.best_model_path

                if args.eval_data_generator is None:
                    _, eval_examples = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.glue_labels,
                        shuffle=True,
                        train_rate=args.train_rate,
                        num_augments=args.num_augments)
                else:
                    _, eval_examples = load_train_val_examples(
                        args,
                        eval_data_generator,
                        self.glue_labels,
                        shuffle=False,
                        train_rate=0.0,
                        num_augments=0)

                model = self.load_model()
                trainer.evaluate(args, model, eval_examples)

            def do_predict(args):

                args.model_path = args.best_model_path
                test_examples = load_test_examples(args, test_data_generator)
                model = self.load_model()
                trainer.predict(args, model, test_examples)
                from .glue_utils import save_glue_preds
                save_glue_preds(args,
                                trainer.pred_results,
                                test_examples,
                                probs=trainer.pred_probs)

            if args.do_train:
                do_train(args)

            elif args.do_eval:
                do_eval(args)

            elif args.do_predict:
                do_predict(args)

            elif args.do_experiment:
                #  import mlflow
                #  from .utils import log_global_params
                #  if args.tracking_uri:
                #      mlflow.set_tracking_uri(args.tracking_uri)
                #  mlflow.set_experiment(args.experiment_name)
                #
                #  with mlflow.start_run(run_name=f"{args.local_id}") as mlrun:
                #      log_global_params(args, self.experiment_params)
                if True:

                    # ----- Train -----
                    do_train(args)

                    # ----- Predict -----
                    do_predict(args)

                    # ----- Submit -----
                    do_submit(args)


class MultiLabelsApp(GlueApp):
    def get_trainer(self):
        # -------------------- Model --------------------
        if self.trainer is None:
            args = self.args

            from .glue.trainer_multilabels import GlueTrainer

            class AppTrainer(GlueTrainer):
                def __init__(self, args, glue_labels):
                    super(AppTrainer, self).__init__(args,
                                                     glue_labels,
                                                     build_model=None)

                #  def on_predict_end(self, args, test_dataset):
                #      super(Trainer, self).on_predict_end(args, test_dataset)

            self.trainer = AppTrainer(args, self.glue_labels)

        return self.trainer


class NerApp:
    def __init__(self,
                 experiment_params,
                 ner_labels,
                 ner_connections,
                 add_special_args=None):
        self.experiment_params = experiment_params
        self.ner_labels = ner_labels
        self.ner_connections = ner_connections

        if experiment_params.ner_params.ner_type == 'span':
            from .ner_span import load_model, get_args, NerTrainer
        elif experiment_params.ner_params.ner_type == 'pn':
            from .ner_pn import load_model, get_args, NerTrainer
        else:
            from .ner import load_model, get_args, NerTrainer

        args = get_args(experiment_params=experiment_params,
                        special_args=[add_special_args])
        args.app_type = 'ner'
        self.args = args
        logger.info(f"args: {args}")
        #  print("[bold cyan]args:[/bold cyan]", args)

        self.trainer = None

    def get_trainer(self):
        # -------------------- Model --------------------
        if self.trainer is None:
            args = self.args
            if args.ner_type == 'span':
                from .ner_span import NerTrainer
            elif args.ner_type == 'pn':
                from .ner_pn import NerTrainer
            else:
                from .ner import NerTrainer

            class AppTrainer(NerTrainer):
                def __init__(self, args, ner_labels):
                    super(AppTrainer, self).__init__(args,
                                                     ner_labels,
                                                     build_model=None)

                #  def on_predict_end(self, args, test_dataset):
                #      super(Trainer, self).on_predict_end(args, test_dataset)

            self.trainer = AppTrainer(args, self.ner_labels)

        return self.trainer

    def load_model(self):
        if self.experiment_params.ner_params.ner_type == 'span':
            from .ner_span import load_model as global_load_model
        elif self.experiment_params.ner_params.ner_type == 'pn':
            from .ner_pn import load_model as global_load_model
        else:
            from .ner import load_model as global_load_model

        model = global_load_model(self.args)
        return model

    def run(self,
            train_data_generator,
            test_data_generator,
            generate_submission=None,
            eval_data_generator=None):
        args = self.args

        assert train_data_generator is not None
        assert test_data_generator is not None

        #  if eval_data_generator is None:
        #      eval_data_generator = train_data_generator

        def do_eda(args):
            from .ner_utils import show_ner_datainfo
            show_ner_datainfo(self.ner_labels, train_data_generator,
                              args.train_file, test_data_generator,
                              args.test_file)

        def do_submit(args):
            from .utils import archive_local_model
            archive_local_model(args)
            submission_file = None
            if generate_submission:
                submission_file = generate_submission(args)

        def do_generate_submission(args):
            if generate_submission is not None:
                if args.dataset_file is None:
                    logger.error(f"dataset_file parameter must be set.")
                    return
                if args.submission_file is None:
                    logger.error(f"submission_file parameter must be set.")
                    return
                generate_submission(args,
                                    reviews_file=args.dataset_file,
                                    submission_file=args.submission_file)

        if args.do_new:
            do_new(args)

        if args.do_eda:
            do_eda(args)

        elif args.do_submit:
            do_submit(args)

        elif args.generate_submission:
            do_generate_submission(args)

        elif args.to_train_poplar:
            from .ner_utils import to_train_poplar
            to_train_poplar(args,
                            train_data_generator,
                            ner_labels=ner_labels,
                            ner_connections=ner_connections,
                            start_page=args.start_page,
                            max_pages=args.max_pages)

        elif args.to_reviews_poplar:
            from .ner_utils import to_reviews_poplar
            to_reviews_poplar(args,
                              ner_labels=ner_labels,
                              ner_connections=ner_connections,
                              start_page=args.start_page,
                              max_pages=args.max_pages)

        else:

            trainer = self.get_trainer()

            from .ner_utils import load_train_val_examples, load_test_examples

            def do_train(args):
                if eval_data_generator is None:
                    train_examples, val_examples = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.ner_labels,
                        shuffle=True,
                        train_rate=args.train_rate,
                        num_augments=args.num_augments,
                        aug_train_only=args.aug_train_only)
                else:
                    train_examples, _ = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.ner_labels,
                        shuffle=True,
                        train_rate=1.0,
                        num_augments=args.num_augments,
                        aug_train_only=args.aug_train_only)

                    _, val_examples = load_train_val_examples(
                        args,
                        eval_data_generator,
                        self.ner_labels,
                        shuffle=False,
                        train_rate=0.0,
                        num_augments=0,
                        aug_train_only=args.aug_train_only)
                trainer.train(args, train_examples, val_examples)

            def do_eval(args):
                args.model_path = args.best_model_path

                if eval_data_generator is None:
                    _, eval_examples = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.ner_labels,
                        shuffle=True,
                        train_rate=args.train_rate,
                        num_augments=args.num_augments,
                        aug_train_only=args.aug_train_only)
                else:
                    _, eval_examples = load_train_val_examples(
                        args,
                        eval_data_generator,
                        self.ner_labels,
                        shuffle=False,
                        train_rate=0.0,
                        num_augments=0,
                        aug_train_only=args.aug_train_only)

                model = self.load_model()
                trainer.evaluate(args, model, eval_examples)

            def do_predict(args):

                from .ner_utils import save_ner_preds
                args.model_path = args.best_model_path
                test_examples = load_test_examples(args, test_data_generator)
                model = self.load_model()
                trainer.predict(args, model, test_examples)
                reviews_file, category_mentions_file = save_ner_preds(
                    args, trainer.pred_results, test_examples)
                return reviews_file, category_mentions_file

            if args.do_train:
                do_train(args)

            if args.do_eval:
                do_eval(args)

            if args.do_predict:
                do_predict(args)

            if args.do_experiment:
                #  import mlflow
                #  from .utils import log_global_params
                #  if args.tracking_uri:
                #      mlflow.set_tracking_uri(args.tracking_uri)
                #  mlflow.set_experiment(args.experiment_name)
                #
                #  with mlflow.start_run(run_name=f"{args.local_id}") as mlrun:
                #      log_global_params(args, self.experiment_params)
                if True:

                    # ----- Train -----
                    do_train(args)

                    # ----- Predict -----
                    do_predict(args)

                    # ----- Submit -----
                    do_submit(args)


class SpoApp:
    def __init__(self,
                 experiment_params,
                 predicate_labels,
                 add_special_args=None):
        self.experiment_params = experiment_params
        self.predicate_labels = predicate_labels

        from .re_casrel import load_model, get_args, SpoTrainer

        args = get_args(experiment_params=experiment_params,
                        special_args=[add_special_args])
        args.app_type = 'spo'
        self.args = args
        logger.info(f"args: {args}")
        #  print("[bold cyan]args:[/bold cyan]", args)

        self.trainer = None

    def get_trainer(self):
        # -------------------- Model --------------------
        if self.trainer is None:
            args = self.args
            from .re_casrel import SpoTrainer

            class AppTrainer(SpoTrainer):
                def __init__(self, args, predicate_labels):
                    super(AppTrainer, self).__init__(args,
                                                     predicate_labels,
                                                     build_model=None)

                #  def on_predict_end(self, args, test_dataset):
                #      super(Trainer, self).on_predict_end(args, test_dataset)

            self.trainer = AppTrainer(args, self.predicate_labels)

        return self.trainer

    def load_model(self):
        from .re_casrel import load_model as global_load_model

        model = global_load_model(self.args)
        return model

    def run(self,
            train_data_generator,
            test_data_generator,
            generate_submission=None,
            eval_data_generator=None):
        args = self.args

        assert train_data_generator is not None
        assert test_data_generator is not None

        #  if eval_data_generator is None:
        #      eval_data_generator = train_data_generator

        def do_eda(args):
            from .spo_utils import show_spo_datainfo
            show_spo_datainfo(self.predicate_labels, train_data_generator,
                              args.train_file, test_data_generator,
                              args.test_file)

        def do_submit(args):
            submission_file = None
            if generate_submission:
                submission_file = generate_submission(args)
            from .utils import archive_local_model
            archive_local_model(args, submission_file)

        if args.do_new:
            do_new(args)

        elif args.do_eda:
            do_eda(args)

        elif args.do_submit:
            do_submit(args)

        else:

            trainer = self.get_trainer()

            from .spo_utils import load_train_val_examples, load_test_examples

            def do_train(args):
                if eval_data_generator is None:
                    train_examples, val_examples = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.predicate_labels,
                        shuffle=True,
                        train_rate=args.train_rate,
                        num_augments=args.num_augments)
                else:
                    train_examples, _ = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.predicate_labels,
                        shuffle=True,
                        train_rate=1.0,
                        num_augments=args.num_augments)

                    _, val_examples = load_train_val_examples(
                        args,
                        eval_data_generator,
                        self.predicate_labels,
                        shuffle=False,
                        train_rate=0.0,
                        num_augments=0)

                trainer.train(args, train_examples, val_examples)

            def do_eval(args):
                args.model_path = args.best_model_path

                if eval_data_generator is None:
                    _, eval_examples = load_train_val_examples(
                        args,
                        train_data_generator,
                        self.predicate_labels,
                        shuffle=True,
                        train_rate=args.train_rate,
                        num_augments=args.num_augments)
                else:
                    _, eval_examples = load_train_val_examples(
                        args,
                        eval_data_generator,
                        self.predicate_labels,
                        shuffle=False,
                        train_rate=0.0,
                        num_augments=0)

                model = self.load_model()
                trainer.evaluate(args, model, eval_examples)

            def do_predict(args):

                args.model_path = args.best_model_path
                test_examples = load_test_examples(args, test_data_generator)
                model = self.load_model()
                trainer.predict(args, model, test_examples)
                from .spo_utils import save_spo_preds
                save_spo_preds(args, trainer.pred_results, test_examples)

            if args.do_train:
                do_train(args)

            elif args.do_eval:
                do_eval(args)

            elif args.do_predict:
                do_predict(args)

            elif args.do_experiment:
                #  import mlflow
                #  from .utils import log_global_params
                #  if args.tracking_uri:
                #      mlflow.set_tracking_uri(args.tracking_uri)
                #  mlflow.set_experiment(args.experiment_name)
                #
                #  with mlflow.start_run(run_name=f"{args.local_id}") as mlrun:
                #      log_global_params(args, self.experiment_params)
                if True:

                    # ----- Train -----
                    do_train(args)

                    # ----- Predict -----
                    do_predict(args)

                    # ----- Submit -----
                    do_submit(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
from collections import Counter

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from loguru import logger
from torch.nn import CrossEntropyLoss
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler
from tqdm import tqdm

from transformers import (AutoConfig, AutoModelForTokenClassification,
                          BertConfig, BertTokenizer, BertTokenizerFast)
#  from transformers.modeling_bert import BertModel, BertPreTrainedModel
from transformers.models.bert import BertModel, BertPreTrainedModel

from ...losses import DiceLoss, FocalLoss, LabelSmoothingCrossEntropy
from ...utils.multiprocesses import (barrier_leader_process,
                                     barrier_member_processes,
                                     is_multi_processes)
from ..models.linears import PoolerEndLogits, PoolerStartLogits
from ..trainer import Trainer, get_default_optimizer_parameters
from .utils import CNerTokenizer, SeqEntityScore, get_entities

#  import mlflow


class BertSpanForNer(BertPreTrainedModel):
    def __init__(self, config):
        super(BertSpanForNer, self).__init__(config)
        self.soft_label = config.soft_label
        self.num_labels = config.num_labels
        self.loss_type = config.loss_type
        self.focalloss_gamma = config.focalloss_gamma
        self.focalloss_alpha = config.focalloss_alpha
        self.diceloss_weight = config.diceloss_weight

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.start_fc = PoolerStartLogits(config.hidden_size, self.num_labels)
        if self.soft_label:
            self.end_fc = PoolerEndLogits(config.hidden_size + self.num_labels,
                                          self.num_labels)
        else:
            self.end_fc = PoolerEndLogits(config.hidden_size + 1,
                                          self.num_labels)
        self.init_weights()


#  class BertSpanForNer:
#      def __init__(self, args):
#          self.args = args
#          self.soft_label = args.soft_label
#          self.num_labels = args.num_labels
#          self.loss_type = args.loss_type
#          self.focalloss_gamma = args.focalloss_gamma
#          self.focalloss_alpha = args.focalloss_alpha
#
#          config_class = AutoConfig
#          model_class = AutoConfig
#          config = config_class.from_pretrained(
#              args.model_path,
#              num_labels=args.num_labels,
#              label2id = args.label2id,
#              id2label = args.id2label,
#              cache_dir=args.cache_dir if args.cache_dir else None,
#          )
#          logger.info(f"model_path: {args.model_path}")
#          logger.info(f"config:{config}")
#          self.bert = model_class.from_pretrained(
#              args.model_path,
#              from_tf=bool(".ckpt" in args.model_path),
#              config=config,
#              cache_dir=args.cache_dir if args.cache_dir else None,
#          )
#
#          self.dropout = nn.Dropout(config.hidden_dropout_prob)
#          self.start_fc = PoolerStartLogits(config.hidden_size, self.num_labels)
#          if self.soft_label:
#              self.end_fc = PoolerEndLogits(config.hidden_size + self.num_labels,
#                                            self.num_labels)
#          else:
#              self.end_fc = PoolerEndLogits(config.hidden_size + 1,
#                                            self.num_labels)
#          self.init_weights()

    def forward(self,
                input_ids,
                attention_mask=None,
                token_type_ids=None,
                start_positions=None,
                end_positions=None):
        #  subjects_ids=None):
        outputs = self.bert(input_ids=input_ids,
                            attention_mask=attention_mask,
                            token_type_ids=token_type_ids)
        sequence_output = outputs[0]
        sequence_output = self.dropout(sequence_output)
        start_logits = self.start_fc(sequence_output)

        if start_positions is not None and self.training:
            if self.soft_label:
                batch_size = input_ids.size(0)
                seq_len = input_ids.size(1)
                label_logits = torch.FloatTensor(batch_size, seq_len,
                                                 self.num_labels)
                label_logits.zero_()
                label_logits = label_logits.to(input_ids.device)
                label_logits.scatter_(2, start_positions.unsqueeze(2), 1)
            else:
                label_logits = start_positions.unsqueeze(2).float()
        else:
            label_logits = F.softmax(start_logits, -1)
            if not self.soft_label:
                label_logits = torch.argmax(label_logits,
                                            -1).unsqueeze(2).float()
        end_logits = self.end_fc(sequence_output, label_logits)
        avg_logits = (start_logits + end_logits) / 2
        outputs = (
            avg_logits,
            start_logits,
            end_logits,
        ) + outputs[2:]

        if start_positions is not None and end_positions is not None:
            assert self.loss_type in [
                'LabelSmoothingCrossEntropy', 'FocalLoss', 'CrossEntropyLoss'
            ]
            if self.loss_type == 'LabelSmoothingCrossEntropy':
                loss_fct = LabelSmoothingCrossEntropy()
            elif self.loss_type == 'FocalLoss':
                loss_fct = FocalLoss(gamma=self.focalloss_gamma,
                                     alpha=self.focalloss_alpha)
            elif self.loss_type == 'DiceLoss':
                loss_fct = DiceLoss(weight=self.diceloss_weight)
            else:
                loss_fct = CrossEntropyLoss()
            start_logits = start_logits.view(-1, self.num_labels)
            end_logits = end_logits.view(-1, self.num_labels)
            active_loss = attention_mask.view(-1) == 1
            active_start_logits = start_logits[active_loss]
            active_end_logits = end_logits[active_loss]

            active_start_labels = start_positions.view(-1)[active_loss]
            active_end_labels = end_positions.view(-1)[active_loss]

            start_loss = loss_fct(active_start_logits, active_start_labels)
            end_loss = loss_fct(active_end_logits, active_end_labels)
            total_loss = (start_loss + end_loss) / 2
            outputs = (total_loss, ) + outputs
            return outputs
        else:
            #  return (0.0, ) + outputs
            #  logger.warning(
            #      f"start_positions: {start_positions}, end_positions: {end_positions}"
            #  )
            return (torch.tensor(0.0).cuda(), ) + outputs
            #  return outputs


class SpanEntityScore(object):
    def __init__(self, id2label, ignore_categories=None):
        self.id2label = id2label
        self.ignore_categories = ignore_categories
        self.reset()

    def reset(self):
        self.origins = []
        self.founds = []
        self.rights = []

    def compute(self, origin, found, right):
        recall = 0 if origin == 0 else (right / origin)
        precision = 0 if found == 0 else (right / found)
        f1 = 0. if recall + precision == 0 else (2 * precision *
                                                 recall) / (precision + recall)
        return recall, precision, f1

    def result(self):
        class_info = {}
        origin_counter = Counter(
            [f"{x[0]}:{self.id2label[x[0]]}" for x in self.origins])
        found_counter = Counter(
            [f"{x[0]}:{self.id2label[x[0]]}" for x in self.founds])
        right_counter = Counter(
            [f"{x[0]}:{self.id2label[x[0]]}" for x in self.rights])

        total_origin = 0
        total_found = 0
        total_right = 0
        for type_, count in origin_counter.items():
            category = type_.split(':')[1]
            origin = count
            found = found_counter.get(type_, 0)
            right = right_counter.get(type_, 0)
            recall, precision, f1 = self.compute(origin, found, right)
            class_info[type_] = {
                "acc": round(precision, 4),
                'recall': round(recall, 4),
                'f1': round(f1, 4),
                'right': right,
                'found': found,
                'origin': origin
            }
            if self.ignore_categories and category in self.ignore_categories:
                pass
            else:
                total_origin += origin
                total_found += found
                total_right += right
        if self.ignore_categories:
            origin = total_origin
            found = total_found
            right = total_right
        else:
            origin = len(self.origins)
            found = len(self.founds)
            right = len(self.rights)
        recall, precision, f1 = self.compute(origin, found, right)
        return {
            'acc': precision,
            'recall': recall,
            'f1': f1,
            'right': right,
            'found': found,
            'origin': origin
        }, class_info

    def update(self, true_subject, pred_subject):
        self.origins.extend(true_subject)
        self.founds.extend(pred_subject)
        self.rights.extend([
            pre_entity for pre_entity in pred_subject
            if pre_entity in true_subject
        ])


MODEL_CLASSES = {
    'bert': (BertConfig, BertSpanForNer, CNerTokenizer),
    #  'bert': (BertConfig, BertSpanForNer, BertTokenizer),
    #  'bert': (BertConfig, BertSpanForNer, BertTokenizerFast),
}


def load_pretrained_tokenizer(args):
    #  config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    #  tokenizer = tokenizer_class.from_pretrained(
    #      args.model_path,
    #      do_lower_case=args.do_lower_case,
    #      is_english=args.is_english,
    #      cache_dir=args.cache_dir if args.cache_dir else None,
    #  )
    from ..token_utils import HFTokenizer
    tokenizer = HFTokenizer(os.path.join(args.model_path, 'vocab.txt'),
                            lowercase=args.do_lower_case,
                            cc=args.cc)

    return tokenizer


def load_pretrained_model(args):
    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_member_processes(args)

    config_class, model_class, tokenizer_class = MODEL_CLASSES[args.model_type]
    config = config_class.from_pretrained(
        args.model_path,
        num_labels=args.num_labels,
        loss_type=args.loss_type,
        cache_dir=args.cache_dir if args.cache_dir else None,
        soft_label=args.soft_label,
    )
    setattr(config, 'label2id', args.label2id)
    setattr(config, 'id2label', args.id2label)
    setattr(config, 'soft_label', args.soft_label)
    setattr(config, 'loss_type', args.loss_type)
    setattr(config, 'focalloss_gamma', args.focalloss_gamma)
    setattr(config, 'focalloss_alpha', args.focalloss_alpha)
    setattr(config, 'diceloss_weight', args.diceloss_weight)
    logger.info(f"model_path: {args.model_path}")
    logger.info(f"config:{config}")
    model = model_class.from_pretrained(
        args.model_path,
        from_tf=bool(".ckpt" in args.model_path),
        config=config,
        cache_dir=args.cache_dir if args.cache_dir else None,
    )

    # make sure only the first process in distributed training
    # will download model & vocab
    barrier_leader_process(args)

    return model


#  def batch_to_input_data(batch):
#      all_input_ids = torch.stack([e.input_ids for e in batch])
#      all_input_mask = torch.stack([e.input_mask for e in batch])
#      all_segment_ids = torch.stack([e.segment_ids for e in batch])
#      all_input_lens = torch.stack([e.input_len for e in batch])
#      all_token_offsets = torch.stack([e.token_offsets for e in batch])
#      all_start_ids = torch.stack([e.start_ids for e in batch])
#      all_end_ids = torch.stack([e.end_ids for e in batch])
#      all_subjects = [e.subjects for e in batch]
#
#      return (all_input_ids, all_input_mask, all_segment_ids, all_start_ids,
#              all_end_ids, all_input_lens, all_subjects, all_token_offsets)


def collate_fn(batch):
    """
    batch should be a list of (sequence, target, length) tuples...
    Returns a padded tensor of sequences sorted from longest to shortest,
    """
    #  all_input_ids, all_input_mask, all_segment_ids, all_start_ids, all_end_ids, all_input_lens = map(
    #      torch.stack, zip(*batch))
    #  from .dataset import batch_to_input_data
    #  all_input_ids, all_input_mask, all_segment_ids, all_start_ids, all_end_ids, all_input_lens, all_subjects, all_token_offsets = batch_to_input_data(
    #      batch)
    all_input_ids = torch.stack([e.input_ids for e in batch])
    all_attention_mask = torch.stack([e.attention_mask for e in batch])
    all_token_type_ids = torch.stack([e.token_type_ids for e in batch])
    all_input_lens = torch.stack([e.input_len for e in batch])
    all_token_offsets = torch.stack([e.token_offsets for e in batch])

    all_start_ids = torch.stack([e.start_ids for e in batch])
    all_end_ids = torch.stack([e.end_ids for e in batch])
    all_subjects = [e.subjects for e in batch]

    max_len = max(all_input_lens).item()
    all_input_ids = all_input_ids[:, :max_len]
    all_attention_mask = all_attention_mask[:, :max_len]
    all_token_type_ids = all_token_type_ids[:, :max_len]
    all_token_offsets = all_token_offsets[:, :max_len]

    all_start_ids = all_start_ids[:, :max_len]
    all_end_ids = all_end_ids[:, :max_len]

    return all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets, all_start_ids, all_end_ids, all_subjects


def bert_extract_item(start_logits,
                      end_logits,
                      lens,
                      confidence,
                      overlap=False):
    num_tokens = lens[0]
    S = []

    starts = torch.argmax(start_logits, -1).cpu().numpy()[0]  #[1:-1]
    ends = torch.argmax(end_logits, -1).cpu().numpy()[0]  #[1:-1]

    # start_logits.shape: (1, sentence_length, num_labels)
    #  logger.debug(f"start_logits: {start_logits.shape}")
    #  start_max_probs = [
    #      f"{start_logits[0][i][x]:.4f}" for i, x in enumerate(starts)
    #  ]
    #  logger.debug(f"start_max_probs: {start_max_probs}")
    #  logger.debug(f"starts: {starts.shape} {starts}")
    starts = [
        x if start_logits[0][i][x] >= confidence else 0
        for i, x in enumerate(starts)
    ]
    ends = [
        x if end_logits[0][i][x] >= confidence else 0
        for i, x in enumerate(ends)
    ]

    #  starts = [starts[0]] + [ x if x != starts[i] else 0 for i, x in enumerate(starts[1:])]
    #  ends = [ends[0]] + [ x if x != ends[i] else 0 for i, x in enumerate(ends[1:])]
    def filter_process(starts):
        new_starts = []
        for i, x in enumerate(starts):
            is_dup = False
            if i < len(starts) - 1 and x == starts[i + 1]:
                is_dup = True
            elif i > 0 and starts[i - 1] == x:
                is_dup = True
            if is_dup:
                new_starts.append(0)
            else:
                new_starts.append(x)
        return new_starts

    #  starts = filter_process(starts)
    #  ends = filter_process(ends)

    #  starts = np.array([x for x in starts if x >= 0 and x < num_tokens])
    #  ends = np.array([x for x in ends if x >= 0 and x < num_tokens])
    starts = starts[:num_tokens]
    ends = ends[:num_tokens]

    #  logger.info(f"start_pred: {starts}")
    #  logger.info(f"end_pred: {ends}")
    #  for i, s_l in enumerate(starts):
    #      if s_l == 0:
    #          continue
    #      for j, e_l in enumerate(ends[i:]):
    #          if s_l == e_l:
    #              S.append((s_l, i, i + j))
    #              break
    #          if i + j < len(starts) - 1 and starts[i + j + 1] != 0:
    #              break
    #  for i in range(len(starts) - 1):
    last_j = -1
    for i in range(len(starts)):
        if i <= last_j:
            continue
        s_l = starts[i]
        if s_l == 0:
            continue
        for j, e_l in enumerate(ends[i:]):
            if s_l == e_l:
                if not overlap:
                    #  if sum(starts[i + 1:i + j + 1]) == 0:
                    S.append((int(s_l), i, i + j))
                    last_j = j
                    i = j + 1
                    break
                else:
                    if sum(starts[i + 1:i + j + 1]) != 0:
                        break
                    S.append((int(s_l), i, i + j))
            if i + j < len(starts) - 1 and starts[i + j + 1] != 0:
                break
    #  S = [x for x in S if x[1] <= x[2]]

    #  for x in S:
    #      assert x[1] >= 0 and x[2] >= 0 and x[1] <= x[2], f"S: {S}"

    return S


def load_model(args):
    model = load_pretrained_model(args)
    model.to(args.device)
    return model


def build_default_model(args):
    """
    自定义模型
    规格要求返回模型(model)、优化器(optimizer)、调度器(scheduler)三元组。
    """

    # -------- model --------
    model = load_pretrained_model(args)
    model.to(args.device)

    # -------- optimizer --------
    from transformers.optimization import AdamW
    optimizer_parameters = get_default_optimizer_parameters(
        model, args.weight_decay)
    optimizer = AdamW(optimizer_parameters,
                      lr=args.learning_rate,
                      eps=args.adam_epsilon,
                      correct_bias=False)

    # -------- scheduler --------
    from transformers.optimization import get_linear_schedule_with_warmup

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=args.total_steps * args.warmup_rate,
        num_training_steps=args.total_steps)

    return model, optimizer, scheduler


def init_labels(args, labels):

    #  args.ner_labels = ['[unused1]', '[unused2]', '[unused3]'] + labels
    #  args.id2label = {i: label for i, label in enumerate(args.ner_labels)}
    #  args.label2id = {label: i for i, label in enumerate(args.ner_labels)}
    args.ner_labels = ['[unused1]'] + labels
    args.id2label = {i: label for i, label in enumerate(args.ner_labels)}
    args.label2id = {label: i for i, label in enumerate(args.ner_labels)}
    args.num_labels = len(args.label2id)

    #  args.id2label = {i + 100: label for i, label in enumerate(args.ner_labels)}
    #  args.label2id = {label: i + 100 for i, label in enumerate(args.ner_labels)}
    #
    #  args.ner_labels = ['[unused1]'] + labels
    #  args.id2label[0] = '[unused1]'
    #  args.label2id['[unused1]'] = 0
    #  args.num_labels = len(args.label2id)

    logger.info(f"args.label2id: {args.label2id}")
    logger.info(f"args.id2label: {args.id2label}")
    logger.info(f"args.num_labels: {args.num_labels}")


class NerTrainer(Trainer):
    def __init__(self, args, ner_labels, build_model=None, tokenizer=None):
        super(NerTrainer, self).__init__(args)
        init_labels(args, ner_labels)
        if tokenizer:
            self.tokenizer = tokenizer
        else:
            self.tokenizer = load_pretrained_tokenizer(args)

        if build_model is None:
            self.build_model = build_default_model
        else:
            self.build_model = build_model

        self.label2id = args.label2id
        self.collate_fn = collate_fn

    #  def examples_to_dataset(self, examples, max_seq_length):
    #      from .dataset import examples_to_dataset
    #      return examples_to_dataset(examples, self.label2id, self.tokenizer,
    #                                 max_seq_length)

    def encode_examples(self,
                        examples,
                        max_seq_length,
                        enttype_labels=None,
                        epoch=-1):
        from .dataset import encode_examples
        return encode_examples(examples,
                               self.label2id,
                               self.tokenizer,
                               max_seq_length,
                               enttype_labels=enttype_labels,
                               epoch=epoch)

    def batch_to_inputs(self, args, batch, known_labels=True):
        inputs = {
            "input_ids": batch[0],
            "attention_mask": batch[1],
            "start_positions": batch[5],
            "end_positions": batch[6],
        }
        if args.model_type != "distilbert":
            # XLM and RoBERTa don"t use segment_ids
            inputs["token_type_ids"] = (
                batch[2] if args.model_type in ["bert", "xlnet"] else None)

        return inputs

    #  def generate_dataloader(self, args, dataset, batch_size, keep_order=True):
    #
    #      Sampler = SequentialSampler if keep_order else RandomSampler
    #      sampler = DistributedSampler(dataset) if is_multi_processes(
    #          args) else Sampler(dataset)
    #      dataloader = DataLoader(dataset,
    #                              sampler=sampler,
    #                              batch_size=batch_size,
    #                              collate_fn=collate_fn)
    #      return dataloader

    def on_eval_start(self, args, eval_dataset):
        self.metric = SpanEntityScore(args.id2label,
                                      ignore_categories=args.ignore_categories)
        self.eval_logs = []

    #  def on_eval_step(self, args, eval_dataset, step, model, inputs, outputs):
    def on_eval_step(self, args, model, step, batch):
        all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets, all_start_ids, all_end_ids, all_subjects = batch

        eval_loss = 0.0
        num_eval_steps = 0
        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_attention_mask[i].view(1, -1),
                "start_positions": all_start_ids[i].view(1, -1),
                "end_positions": all_end_ids[i].view(1, -1),
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use segment_ids
                inputs["token_type_ids"] = (all_token_type_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)

            outputs = model(**inputs)
            tmp_eval_loss, avg_logits, start_logits, end_logits = outputs[:4]
            #  tmp_eval_loss, start_logits, end_logits = outputs[:3]
            eval_loss += tmp_eval_loss
            num_eval_steps += 1

            start_logits = F.softmax(start_logits, -1)
            end_logits = F.softmax(end_logits, -1)
            num_tokens = int(all_input_lens[i])
            R = bert_extract_item(start_logits,
                                  end_logits, [num_tokens],
                                  args.confidence,
                                  overlap=args.allow_overlap)

            T = all_subjects[i]

            #  logger.warning(f"R: {R}")
            #  logger.warning(f"T: {T}")

            self.metric.update(true_subject=T, pred_subject=R)

            def calculate_eval(true_subject, pred_subject):
                origins = true_subject
                founds = pred_subject
                rights = [
                    pre_entity for pre_entity in pred_subject
                    if pre_entity in true_subject
                ]
                origin = len(origins)
                found = len(founds)
                right = len(rights)
                recall = 0 if origin == 0 else (right / origin)
                precision = 0 if found == 0 else (right / found)
                f1 = 0. if recall + precision == 0 else (
                    2 * precision * recall) / (precision + recall)
                return precision, recall, f1

            precision, recall, f1 = calculate_eval(T, R)

            token_offsets = all_token_offsets[i]
            #  label_entities = [[
            #      args.id2label[x[0]], token_offsets[x[1]][0].item(),
            #      token_offsets[x[2]][-1].item() - 1
            #  ] for x in R]
            self.eval_logs.append(
                (num_tokens, R, token_offsets, (precision, recall, f1)))

        eval_loss = eval_loss / num_eval_steps
        eval_info, entity_info = self.metric.result()
        results = {f'{key}': value for key, value in eval_info.items()}
        results['loss'] = eval_loss

        #  if args.do_experiment:
        #      mlflow.log_metric('loss', eval_loss.item())
        #      for key, value in eval_info.items():
        #          mlflow.log_metric(key, value)

        return (eval_loss, ), results

    def on_eval_end(self, args, eval_examples):
        from ...utils.ner_utils import get_ner_results
        results = get_ner_results(self.metric)

        eval_logs_file = f"{args.latest_dir}/eval_logs_{args.epoch}.json"

        def entity_text(example, token_offsets, m, s, e):
            s = token_offsets[s][0].item()
            e = token_offsets[e][-1].item()
            return f"{args.id2label[m]}|({m}, {s}, {e})|{example.text[s:e]}"

        all_eval_logs = [{
            'text':
            example.text,
            'num_tokens':
            num_tokens,
            'subjects': [
                entity_text(example, token_offsets, m, s, e)
                for m, s, e in example.subjects
            ],
            'right': [
                entity_text(example, token_offsets, m, s, e)
                for m, s, e in sorted(list(set(R) & set(example.subjects)),
                                      key=lambda x: x[1])
            ],
            'miss': [
                entity_text(example, token_offsets, m, s, e)
                for m, s, e in sorted(list(set(example.subjects) - set(R)),
                                      key=lambda x: x[1])
            ],
            'extra': [
                entity_text(example, token_offsets, m, s, e)
                for m, s, e in sorted(list(set(R) - set(example.subjects)),
                                      key=lambda x: x[1])
            ],
            'pred':
            [entity_text(example, token_offsets, m, s, e) for m, s, e in R],
            'acc_recall_f1':
            x
        } for example, (
            num_tokens,
            R, token_offsets,
            x) in tqdm(zip(eval_examples, self.eval_logs), desc="eval_logs")]
        json.dump(all_eval_logs,
                  open(eval_logs_file, 'w'),
                  ensure_ascii=False,
                  indent=2)

        logger.warning(f"Saved eval logs in {eval_logs_file}")

        return results

    def on_predict_start(self, args, test_features):
        self.pred_results = []

    def on_predict_step(self, args, model, step, batch):
        all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets, all_start_ids, all_end_ids, all_subjects = batch

        for i in range(all_input_ids.size()[0]):
            inputs = {
                "input_ids": all_input_ids[i].view(1, -1),
                "attention_mask": all_attention_mask[i].view(1, -1),
            }
            if args.model_type != "distilbert":
                # XLM and RoBERTa don"t use segment_ids
                inputs["token_type_ids"] = (all_token_type_ids[i].view(
                    1, -1) if args.model_type in ["bert", "xlnet"] else None)
            token_offsets = all_token_offsets[i]

            outputs = model(**inputs)
            _, _, start_logits, end_logits = outputs[:4]
            #  _, start_logits, end_logits = outputs[:3]

            start_logits = F.softmax(start_logits, -1)
            end_logits = F.softmax(end_logits, -1)
            num_tokens = int(all_input_lens[i])
            R = bert_extract_item(start_logits,
                                  end_logits, [num_tokens],
                                  args.confidence,
                                  overlap=args.allow_overlap)

            if R:
                label_entities = [[
                    args.id2label[x[0]], token_offsets[x[1]][0].item(),
                    token_offsets[x[2]][-1].item() - 1
                ] for x in R]
                if args.ignore_categories:
                    label_entities = [
                        x for x in label_entities
                        if x[0] not in args.ignore_categories
                    ]
                label_entities = [
                    x for x in label_entities
                    if x[1] <= x[2] and x[1] >= 0 and x[2] >= 0
                ]

            else:
                label_entities = []

            #  logger.debug(f"{label_entities}")
            json_d = {}
            json_d['id'] = step
            #  tag_seq = [args.id2label[x] for x in preds]
            #  json_d['tag_seq'] = " ".join(tag_seq)
            json_d['entities'] = label_entities

            #  logger.debug(f"{json_d}")

            self.pred_results.append(json_d)

    def on_predict_end(self, args, test_dataset):
        return self.pred_results



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import json
import rich
from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import List, Optional, Tuple, Type, Union

from loguru import logger

from transformers import HfArgumentParser
from transformers import TrainingArguments as HfTrainingArguments
from transformers import set_seed
from transformers.trainer_utils import is_main_process


def compare_arguments(one_args, another_args, return_identical=False):
    #  assert isinstance(
    #      another_args, BaseArguments
    #  ), f"another_args {type(another_args)} must be subtype of BaseArguments."

    clsname = one_args.__class__.__name__

    diff_attrs = []
    identical_attrs = []
    left_only_attrs = []
    right_only_attrs = []
    for k, v in one_args.__dict__.items():
        if k not in another_args.__dict__:
            left_only_attrs.append((clsname, k, v))
        else:
            v1 = getattr(another_args, k)
            if v == v1:
                if return_identical:
                    identical_attrs.append((clsname, k, v))
            else:
                if isinstance(v, BaseArguments):
                    comp_result = v.compare(v1, return_identical)
                    diff_attrs.extend(comp_result['diff'])
                    left_only_attrs.extend(comp_result['left_only'])
                    right_only_attrs.extend(comp_result['right_only'])
                    if return_identical:
                        identical_attrs.extend(comp_result['identical'])
                else:
                    diff_attrs.append((clsname, k, v, v1))

    for k, v in another_args.__dict__.items():
        if k not in one_args.__dict__:
            right_only_attrs.append((another_args.__class__.__name__, k, v))

    comp_result = {
        'diff': diff_attrs,
        'left_only': left_only_attrs,
        'right_only': right_only_attrs,
    }
    if return_identical:
        comp_result['identical'] = identical_attrs

    return comp_result


@dataclass
class BaseArguments:
    def to_dict(self):
        return asdict(self)

    def check_args(self):
        pass


@dataclass
class DataArguments(BaseArguments):
    max_length: int = field(
        default=128,
        metadata={
            "help":
            "The maximum total input sequence length after tokenization. "
            "Sequences longer than this will be truncated, sequences shorter will be padded."
        },
    )
    train_file: Optional[str] = field(
        default=None,
        metadata={"help": "The path of train file."},
    )
    test_file: Optional[str] = field(
        default=None,
        metadata={"help": "The path of test file."},
    )
    padding: Optional[str] = field(
        default='max_length',
        metadata={'help': "Pad to ['longest', 'max_length', 'do_not_pad']"})
    pad_to_max_length: Optional[bool] = field(
        default=True, metadata={"help": "Pad to max length (Deprecated)"})

    overwrite_cache: bool = field(
        default=False,
        metadata={"help": ("Overwrite the content of the cache directory.")},
    )
    cache_dir: str = field(
        default=None,
        metadata={"help": "Cache dir"},
    )

    #  split_ratio: Optional[Union[float, List, Tuple]] = field(
    split_ratios: Optional[float] = field(
        default=0.9,
        #  default_factory=list,
        metadata={
            "help":
            "Split ratio of train/eval/test dataset. float for train or [train, eval, test]"
        },
    )
    random_state: Optional[int] = field(
        default=None,
        metadata={"help": "Random state"},
    )
    # NER: 保留未标注实体的拆分后的句子作为训练数据
    preserve_no_entity: bool = field(
        default=False,
        metadata={
            "help": "Preseve no enity segmented sentences for training."
        })

    def __post_init__(self):
        pass

    def check_args(self):
        pass


@dataclass
class ModelArguments(BaseArguments):
    model_name_or_path: str = field(
        default=None,
        metadata={"help": "Model name or path."},
    )
    checkpoint_path: str = field(
        default=None,
        metadata={"help": "Checkpoint path."},
    )
    config_name: str = field(
        default=None,
        metadata={"help": "Pretrained model config name."},
    )
    tokenizer_name: Optional[str] = field(
        default=None,
        metadata={
            "help":
            "Pretrained tokenizer name or path if not the same as model_name"
        })
    #  num_labels: int = field(
    #      default=None,
    #      metadata={"help": "Number of labels"},
    #  )
    use_fast_tokenizer: bool = field(
        default=True,
        metadata={
            "help":
            "Whether to use one of the fast tokenizer (backed by the tokenizers library) or not."
        },
    )
    model_revision: str = field(
        default="main",
        metadata={
            "help":
            "The specific model version to use (can be a branch name, tag name or commit id)."
        },
    )
    use_auth_token: bool = field(
        default=False,
        metadata={
            "help":
            "Will use the token generated when running `transformers-cli login` (necessary to use this script "
            "with private models)."
        },
    )

    def __post_init__(self):
        pass

    def check_args(self):
        pass


@dataclass
class TrainingArguments(BaseArguments):  #HfTrainingArguments):

    theta_version: Optional[str] = field(
        default=None,
        metadata={"help": "Theta version."},
    )
    seed: int = field(
        default=42,
        metadata={
            "help":
            "Random seed that will be set at the beginning of training."
        },
    )
    output_dir: Optional[str] = field(
        default="outputs",
        metadata={
            "help":
            "The output directory where the model predictions and checkpoints will be written."
        },
    )
    submissions_dir: Optional[str] = field(
        default="submissions",
        metadata={
            "help":
            "The output directory where the submission files will be written."
        },
    )
    task_name: Optional[str] = field(
        default=None,
        metadata={"help": "Task name"},
    )
    task_dir: Optional[str] = field(
        default=None,
        metadata={
            "help":
            "The task output directory where the model predictions and checkpoints will be written."
        },
    )
    latest_path: Optional[str] = field(
        default=None,
        metadata={
            "help":
            "The task output directory where the model predictions and checkpoints will be written."
        },
    )
    # -------------------- commands --------------------
    do_train: bool = field(default=False,
                           metadata={"help": "Whether to run training."})
    do_eval: bool = field(
        default=False,
        metadata={"help": "Whether to run eval on the dev set."})
    do_predict: bool = field(
        default=False,
        metadata={"help": "Whether to run predictions on the test set."})

    do_submit: bool = field(default=False,
                            metadata={"help": "Whether to run submit."})

    # -------------------- training --------------------
    max_epochs: int = field(
        default=3,
        metadata={"help": "Total number of training epochs to perform."})
    num_train_epochs: int = field(
        default=None,
        metadata={
            "help": "Total number of training epochs to perform. (Deprecated)"
        },
    )
    max_steps: int = field(default=None, metadata={"help": "Max train steps."})
    max_steps: int = field(
        default=None,
        metadata={
            "help":
            "If > 0: set total number of training steps to perform. Override max_epochs."
        },
    )

    gpus: int = field(default=1, metadata={"help": "gpus"})
    learning_rate: float = field(
        default=2e-5,
        metadata={"help": "The initial learning rate for AdamW."},
    )
    weight_decay: float = field(
        default=0.01,
        metadata={"help": "Weight decay for AdamW if we apply some."},
    )
    per_device_train_batch_size: int = field(
        default=8,
        metadata={"help": "Batch size per GPU/TPU core/CPU for training."},
    )

    per_device_eval_batch_size: int = field(
        default=8,
        metadata={"help": "Batch size per GPU/TPU core/CPU for evaluation."},
    )
    per_device_test_batch_size: int = field(
        default=8,
        metadata={"help": "Batch size per GPU/TPU core/CPU for testing."},
    )

    warmup_method: str = field(
        default='by_epoch',
        metadata={
            "help":
            "Linear methods. ['by_epoch', 'by_steps', 'by_rate', 'auto']"
        },
    )
    warmup_rate: float = field(
        default=0.1,
        metadata={"help": "Linear warmup over warmup_rate."},
    )

    warmup_epochs: int = field(
        default=1,
        metadata={"help": "Linear warmup over warmup_epochs."},
    )
    warmup_steps: int = field(
        default=None,
        metadata={"help": "Linear warmup over warmup_steps."},
    )
    #  val_check_interval: float = field(
    #      default=1.0,
    #      metadata={"help": "val_check_interval."},
    #  )

    eval_steps_or_interval: float = field(
        default=0.5,
        metadata={"help": "Run an evaluation every X steps."},
    )

    #  logging_steps: int = field(
    #      default=None,
    #      metadata={"help": "Log every X updates steps."},
    #  )

    #  save_steps: int = field(
    #      default=None,
    #      metadata={"help": "Save checkpoint every X updates steps."},
    #  )
    metric_for_best_model: Optional[str] = field(
        default="val_loss",
        metadata={
            "help": "The metric to use to compare two different models."
        })
    greater_is_better: Optional[bool] = field(
        default=False,
        metadata={
            "help":
            "Whether the `metric_for_best_model` should be maximized or not."
        })

    save_top_k: int = field(
        default=3,
        metadata={"help": "Save top k best models."},
    )
    save_weights_only: bool = field(
        default=True,
        metadata={"help": "Save best model weights only."},
    )
    fp16: bool = field(
        default=False,
        metadata={
            "help":
            "Whether to use 16-bit (mixed) precision (through NVIDIA Apex) instead of 32-bit"
        },
    )
    earlystopping_patience: int = field(
        default=0,
        metadata={"help": "EarlyStopping patience."},
    )
    gradient_accumulation_steps: int = field(
        default=1,
        metadata={
            "help":
            "Number of updates steps to accumulate before performing a backward/update pass."
        },
    )
    evaluation_strategy: str = field(
        default="no",
        metadata={
            "help":
            "The evaluation strategy to use. One of EvaluationStrategy ['no', 'steps', 'epochs']."
        },
    )

    overwrite_output_dir: bool = field(
        default=False,
        metadata={
            "help":
            ("Overwrite the content of the output directory."
             "Use this to continue training if output_dir points to a checkpoint directory."
             )
        },
    )

    #  run_name: Optional[str] = field(
    #      default=None,
    #      metadata={
    #          "help":
    #          "An optional descriptor for the run. Notably used for wandb logging."
    #      },
    #  )
    #
    #  adam_epsilon: float = field(
    #      default=1e-8,
    #      metadata={"help": "Epsilon for AdamW optimizer."},
    #  )
    #
    #  lr_scheduler_type: str = field(
    #      default="linear",
    #      metadata={
    #          "help":
    #          "The scheduler type to use. One of SchedulerType "
    #          "['linear', 'cosine', 'cosine_with_restarts', 'polynomial', 'constant', 'constant_with_warmup']"
    #      },
    #  )
    #
    #
    #  label_smoothing_factor: float = field(
    #      default=0.0,
    #      metadata={
    #          "help":
    #          "The label smoothing epsilon to apply (zero means no label smoothing)."
    #      },
    #  )
    #
    #  adafactor: bool = field(
    #      default=False,
    #      metadata={"help": "Whether or not to replace AdamW by Adafactor."},
    #  )
    #
    #  group_by_length: bool = field(
    #      default=False,
    #      metadata={
    #          "help":
    #          "Whether or not to group samples of roughly the same length together when batching."
    #      },
    #  )
    #
    #  logging_steps: int = field(
    #      default=500,
    #      metadata={"help": "Log every X updates steps."},
    #  )
    #
    #  eval_steps: int = field(
    #      default=None,
    #      metadata={"help": "Run an evaluation every X steps."},
    #  )
    #
    #  save_steps: int = field(
    #      default=500,
    #      metadata={"help": "Save checkpoint every X updates steps."},
    #  )
    #
    #  save_total_limit: Optional[int] = field(
    #      default=None,
    #      metadata={
    #          "help":
    #          ("Limit the total amount of checkpoints."
    #           "Deletes the older checkpoints in the output_dir. Default is unlimited checkpoints"
    #           )
    #      },
    #  )
    #
    #  load_best_model_at_end: Optional[bool] = field(
    #      default=False,
    #      metadata={
    #          "help":
    #          "Whether or not to load the best model found during training at the end of training."
    #      },
    #  )
    #  metric_for_best_model: Optional[str] = field(
    #      default=None,
    #      metadata={
    #          "help": "The metric to use to compare two different models."
    #      })

    def __post_init__(self):
        pass

    @property
    def task_id(self):
        task_id = os.path.basename(self.task_dir)
        return task_id

    def check_args(self):
        if self.do_train:
            #  if self.max_epochs is None and self.max_steps is None:
            #      logger.error(
            #          f"Either max_epochs or max_steps must be set to not none.")
            #      raise ValueError(
            #          f"Either max_epochs or max_steps must be set to not none.")

            if self.max_epochs is not None:
                self.num_train_epochs = self.max_epochs
            else:
                if self.num_train_epochs is not None:
                    logger.warning(
                        f"Argument num_train_epochs is deprecated, use max_epochs please."
                    )
                    #  self.max_epochs = self.num_train_epochs


def ensure_task_dir(training_args):
    #  if training_args.task_dir is None:
    #      if training_args.task_name is not None:
    #          training_args.task_dir = os.path.join(training_args.output_dir,
    #                                                training_args.task_name)
    #      else:
    #          training_args.task_dir = training_args.output_dir

    #  os.makedirs(training_args.task_dir, exist_ok=True)

    #  task_name = os.path.basename(training_args.task_dir).split('_')[0]
    task_name = training_args.task_name
    latest_path = os.path.join(training_args.output_dir,
                               f"{training_args.task_name}_latest")
    training_args.latest_path = latest_path

    if training_args.do_train:
        ts = datetime.now().strftime("%Y%m%d%H%M%S")
        if training_args.task_dir is None:
            if training_args.task_name is not None:
                training_args.task_dir = os.path.join(
                    training_args.output_dir,
                    f"{training_args.task_name}_{ts}")
            else:
                training_args.task_dir = os.path.join(training_args.output_dir,
                                                      f"task_{ts}")
        os.makedirs(training_args.task_dir, exist_ok=True)

        #  if os.path.islink(latest_path):
        #      os.unlink(latest_path)
        #  os.symlink(os.path.basename(training_args.task_dir), latest_path)
    else:
        if training_args.task_dir is None:

            training_args.task_dir = latest_path
            #  training_args.task_dir = os.path.realpath(latest_path)


@dataclass
class TaskArguments():
    rootdir: str = "outputs"
    data_args: DataArguments = field(
        default=DataArguments(),
        metadata={"help": "Arguments of data."},
    )
    model_args: ModelArguments = field(
        default=ModelArguments(),
        metadata={"help": "Arguments of model."},
    )
    training_args: TrainingArguments = field(
        default=TrainingArguments(),
        metadata={"help": "Arguments of training."},
    )
    remaining_args: list = field(
        default_factory=list,
        metadata={"help": "Remaining arguments come from command-line."},
    )

    def __post_init__(self):
        self.check_args()

    def check_args(self):
        self.data_args.check_args()
        self.model_args.check_args()
        self.training_args.check_args()

    def to_dict(self):
        task_args_dict = self.training_args.to_dict()
        task_args_dict.update(self.model_args.to_dict())
        task_args_dict.update(self.data_args.to_dict())
        return task_args_dict

    def show(self):
        training_args = self.training_args

        # Log on each process the small summary:
        logger.warning(
            f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}"
            +
            f"distributed training: {bool(training_args.local_rank != -1)}, 16-bits training: {training_args.fp16}"
        )
        # Set the verbosity to info of the Transformers logger (on main process only):
        if is_main_process(training_args.local_rank):
            #  transformers.utils.logging.set_verbosity_info()
            transformers.utils.logging.set_verbosity_info()
            transformers.utils.logging.enable_default_handler()
            transformers.utils.logging.enable_explicit_format()
        logger.info(f"Training/evaluation parameters {training_args}")

    @classmethod
    def parse_json_file(cls,
                        json_file,
                        data_args_cls=DataArguments,
                        model_args_cls=ModelArguments,
                        training_args_cls=TrainingArguments):
        logger.info(f"Prase json file {json_file}")
        parser = HfArgumentParser(
            (data_args_cls, model_args_cls, training_args_cls))
        data_args, model_args, training_args = parser.parse_json_file(
            json_file)

        remaining_args = None
        task_args = cls(data_args=data_args,
                        model_args=model_args,
                        training_args=training_args,
                        remaining_args=remaining_args)
        return task_args

    @classmethod
    def parse_args(cls,
                   data_args_cls=DataArguments,
                   model_args_cls=ModelArguments,
                   training_args_cls=TrainingArguments):
        logger.info(f"Start parse_args")
        parser = HfArgumentParser(
            (data_args_cls, model_args_cls, training_args_cls))
        #  (DataArguments, ModelArguments, TrainingArguments))
        #  HfTrainingArguments))  #TrainingArguments))
        remaining_args = None
        if len(sys.argv) == 2 and sys.argv[1].endswith(".json"):
            # If we pass only one argument to the script and it's the path to a json file,
            # let's parse it to get our arguments.
            data_args, model_args, training_args = parser.parse_json_file(
                json_file=os.path.abspath(sys.argv[1]))
        else:
            data_args, model_args, training_args, remaining_args = parser.parse_args_into_dataclasses(
                return_remaining_strings=True)

        assert training_args.output_dir is not None
        os.makedirs(training_args.output_dir, exist_ok=True)

        ensure_task_dir(training_args)

        model_args.checkpoint_path = f"{training_args.latest_path}/checkpoint"

        if data_args.cache_dir is None:
            data_args.cache_dir = f"{training_args.output_dir}/.cache.{training_args.task_name}"
        os.makedirs(data_args.cache_dir, exist_ok=True)

        os.makedirs(training_args.submissions_dir, exist_ok=True)

        set_seed(training_args.seed)

        task_args = cls(data_args=data_args,
                        model_args=model_args,
                        training_args=training_args,
                        remaining_args=remaining_args)

        task_log_file = os.path.join(training_args.task_dir, "task.log")
        logger.add(task_log_file)
        #  logger.info(f"task_args: {task_args}")
        rich.print(f"task_args: {task_args}")

        if training_args.do_train:
            task_args_file = os.path.join(training_args.task_dir,
                                          "train_task_args.json")
            json.dump(asdict(task_args),
                      open(task_args_file, 'w'),
                      ensure_ascii=False,
                      indent=2)
        elif training_args.do_predict:
            task_args_file = os.path.join(training_args.task_dir,
                                          "predict_task_args.json")
            json.dump(asdict(task_args),
                      open(task_args_file, 'w'),
                      ensure_ascii=False,
                      indent=2)

        return task_args

    @classmethod
    def merge_args(cls, args, task_args):
        from dataclasses import asdict
        for k, v in asdict(task_args).items():
            setattr(args, k, v)
        return args


def generate_method_kwargs_from_arguments(cls, method, args: dict):
    import inspect
    from dataclasses import asdict
    valid_kwargs = inspect.signature(cls.__dict__[method]).parameters
    kwargs = dict((name, args[name]) for name in valid_kwargs if name in args)
    return kwargs


def create_instance_from_arguments(cls, args: dict):
    kwargs = generate_method_kwargs_from_arguments(cls,
                                                   method="__init__",
                                                   args=args)
    return cls(**kwargs)


if __name__ == '__main__':
    task_args = TaskArguments.parse_args()
    logger.debug(f"{task_args}")
    logger.info(f"{asdict(task_args)}")

    default_args = TaskArguments()
    comp_result = task_args.compare(default_args)
    logger.warning(f"{comp_result}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class BaseTagger:
    def __init__(self, tokenizer, label2id):
        self.tokenizer = tokenizer
        self.label2id = label2id

    def encode(self, examples, *kwarg):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tqdm import tqdm
from loguru import logger


class BaseModel:
    def __init__(self):
        pass

    def load(self, model_path):
        raise NotImplementedError

    def save(self, model_path):
        raise NotImplementedError

    def train(self, train_examples, eval_examples):
        raise NotImplementedError

    def evaluate(self, eval_examples):
        raise NotImplementedError

    def predicgt(self, test_examples):
        raise NotImplementedError


class TransformerModel(BaseModel):
    def __init__(self, args, tagger, trainer, extractor):
        super(TransformerModel, self).__init__()
        self.args = args
        self.tagger = tagger
        self.trainer = trainer
        self.extractor = extractor

    def load_model(self, model_path=None):
        if self.trainer is not None:
            return self.trainer.load_model(model_path)
        return None

    def save_model(self, model_path):
        pass

    def train(self, train_examples, eval_examples):
        train_data = self.tagger.encode(train_examples)
        eval_data = self.tagger.encode(eval_examples)

    def evaluate(self, eval_examples):
        raise NotImplementedError

    def predicgt(self, test_examples):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
from typing import Callable, List, Tuple, Union

from loguru import logger
from tqdm import tqdm


class BaseSample:
    def __init__(self,
                 data_file: str = None,
                 data_generator: Callable = None,
                 data_list: List = None,
                 labels_list: List = None):
        self.data_list = data_list if data_list is not None else []
        self.labels_list = labels_list if labels_list is not None else []

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx: int):
        return self.data_list[idx]

    def _new_instance(self, **kwargs):
        raise NotImplementedError

    def shuffle(self, random_state: int = None):
        if random_state is not None:
            random.seed(random_state)
        random.shuffle(self.data_list)

        return self

    def split(self, ratios: Union[float, List, Tuple] = None):
        if type(ratios) == float:
            train_rate = ratios
            eval_rate = 1.0 - train_rate
            test_rate = 0.0
        elif type(ratios) == list or type(ratios) == tuple:
            if len(ratios) == 2:
                train_rate, eval_rate = ratios
                test_rate = 1.0 - train_rate - eval_rate
            elif len(ratios) == 3:
                train_rate, eval_rate, test_rate = ratios
        else:
            raise ValueError(f"ratios: {ratios} must be float or list.")

        num_total_samples = self.__len__()
        num_train_samples = int(num_total_samples * train_rate)
        if test_rate > 0.0:
            num_eval_samples = int(num_total_samples * eval_rate)
            num_test_samples = num_total_samples - num_train_samples - num_eval_samples
        else:
            num_eval_samples = num_total_samples - num_train_samples
            num_test_samples = 0

        train_samples = self._new_instance(
            data_list=self.data_list[:num_train_samples],
            labels_list=self.labels_list)
        eval_samples = self._new_instance(
            data_list=self.data_list[num_train_samples:num_train_samples +
                                     num_eval_samples],
            labels_list=self.labels_list)
        if num_test_samples > 0:
            test_samples = self._new_instance(
                data_list=self.data_list[-num_test_samples:],
                labels_list=self.labels_list)
            return train_samples, eval_samples, test_samples
        else:
            return train_samples, eval_samples


class GlueSample(BaseSample):
    def __init__(self, **kwargs):
        super(GlueSample, self).__init__(**kwargs)

    def _new_instance(self, **kwargs):
        return GlueSample(**kwargs)

    @property
    def num_labels(self):
        if self.labels_list is None:
            return 0
        else:
            return len(self.labels_list)

    @classmethod
    def from_generator(cls, data_generator: Callable, data_source: str = None):
        samples = GlueSample()
        labels_list = []
        for sid, text_a, text_b, label in data_generator(data_source):
            if label not in labels_list:
                labels_list.append(label)
            samples.data_list.append((sid, text_a, text_b, label))

        labels_list = sorted(labels_list)
        logger.info(f"Loaded labels: {labels_list}")
        if not samples.labels_list:
            samples.labels_list = labels_list

        return samples


class EntitySample(BaseSample):
    def __init__(self, **kwargs):
        super(EntitySample, self).__init__(**kwargs)

    def _new_instance(self, **kwargs):
        return EntitySample(**kwargs)

    @classmethod
    def from_generator(cls, data_generator: Callable, data_source: str = None):
        samples = EntitySample()
        categories = []
        for guid, text, _, json_tags in data_generator(data_source):
            tagged_text = TaggedText(guid, text)
            for json_tag in json_tags:
                entity_tag = EntityTag().from_dict(json_tag)
                tagged_text.add_tag(entity_tag)

                category = entity_tag.category
                if category not in categories:
                    categories.append(category)

            samples.data_list.append(tagged_text)

        logger.info(f"Loaded categories: {categories}")
        if not samples.labels_list:
            samples.labels_list = categories

        return samples


class SPOSample(BaseSample):
    def __init__(self, **kwargs):
        super(SPOSample, self).__init__(**kwargs)

    def _new_instance(self, **kwargs):
        return SPOSample(**kwargs)



# -*- coding: utf-8 -*-
# File: parallel_map.py
import copy
import ctypes
import multiprocessing as mp
import numpy as np
import threading
import zmq
from six.moves import queue

from ..utils.concurrency import StoppableThread, enable_death_signal
from ..utils.serialize import dumps_once as dumps, loads_once as loads
from .base import DataFlow, DataFlowReentrantGuard, ProxyDataFlow
from .common import RepeatedData, BatchData
from .parallel import _bind_guard, _get_pipe_name, _MultiProcessZMQDataFlow, _repeat_iter, _zmq_catch_error

__all__ = ['MultiThreadMapData',
           'MultiProcessMapData', 'MultiProcessMapDataZMQ',
           'MultiProcessMapAndBatchData', 'MultiProcessMapAndBatchDataZMQ']


class _ParallelMapData(ProxyDataFlow):
    def __init__(self, ds, buffer_size, strict=False):
        super(_ParallelMapData, self).__init__(ds)
        assert buffer_size > 0, buffer_size
        self._buffer_size = buffer_size
        self._buffer_occupancy = 0  # actual #elements in buffer, only useful in strict mode
        self._strict = strict

    def reset_state(self):
        super(_ParallelMapData, self).reset_state()
        if not self._strict:
            ds = RepeatedData(self.ds, -1)
        else:
            ds = self.ds
        self._iter = ds.__iter__()

    def _recv(self):
        pass

    def _send(self, dp):
        pass

    def _recv_filter_none(self):
        ret = self._recv()
        assert ret is not None, \
            "[{}] Map function cannot return None when strict mode is used.".format(type(self).__name__)
        return ret

    def _fill_buffer(self, cnt=None):
        if cnt is None:
            cnt = self._buffer_size - self._buffer_occupancy
        try:
            for _ in range(cnt):
                dp = next(self._iter)
                self._send(dp)
        except StopIteration:
            raise RuntimeError(
                "[{}] buffer_size cannot be larger than the size of the DataFlow when strict=True! "
                "Please use a smaller buffer_size!".format(type(self).__name__))
        self._buffer_occupancy += cnt

    def get_data_non_strict(self):
        for dp in self._iter:
            self._send(dp)
            ret = self._recv()
            if ret is not None:
                yield ret

    def get_data_strict(self):
        self._fill_buffer()
        for dp in self._iter:
            self._send(dp)
            yield self._recv_filter_none()
        self._iter = self.ds.__iter__()   # refresh

        # first clear the buffer, then fill
        for k in range(self._buffer_size):
            dp = self._recv_filter_none()
            self._buffer_occupancy -= 1
            if k == self._buffer_size - 1:
                self._fill_buffer()
            yield dp

    def __iter__(self):
        if self._strict:
            yield from self.get_data_strict()
        else:
            yield from self.get_data_non_strict()


class MultiThreadMapData(_ParallelMapData):
    """
    Same as :class:`MapData`, but start threads to run the mapping function.
    This is useful when the mapping function is the bottleneck, but you don't
    want to start processes for the entire dataflow pipeline.

    The semantics of this class is **identical** to :class:`MapData` except for the ordering.
    Threads run in parallel and can take different time to run the
    mapping function. Therefore the order of datapoints won't be preserved.

    When ``strict=True``, ``MultiThreadMapData(df, ...)``
    is guaranteed to produce the exact set of data as ``MapData(df, ...)``,
    if both are iterated until ``StopIteration``. But the produced data will have different ordering.
    The behavior of strict mode is undefined if the given dataflow ``df`` is infinite.

    When ``strict=False``, the data that's produced by ``MultiThreadMapData(df, ...)``
    is a reordering of the data produced by ``RepeatedData(MapData(df, ...), -1)``.
    In other words, first pass of ``MultiThreadMapData.__iter__`` may contain
    datapoints from the second pass of ``df.__iter__``.


    Note:
        1. You should avoid starting many threads in your main process to reduce GIL contention.

           The threads will only start in the process which calls :meth:`reset_state()`.
           Therefore you can use ``MultiProcessRunnerZMQ(MultiThreadMapData(...), 1)``
           to reduce GIL contention.
    """
    class _Worker(StoppableThread):
        def __init__(self, inq, outq, evt, map_func):
            super(MultiThreadMapData._Worker, self).__init__(evt)
            self.inq = inq
            self.outq = outq
            self.func = map_func
            self.daemon = True

        def run(self):
            try:
                while True:
                    dp = self.queue_get_stoppable(self.inq)
                    if self.stopped():
                        return
                    # cannot ignore None here. will lead to unsynced send/recv
                    obj = self.func(dp)
                    self.queue_put_stoppable(self.outq, obj)
            except Exception:
                if self.stopped():
                    pass        # skip duplicated error messages
                else:
                    raise
            finally:
                self.stop()

    def __init__(self, ds, num_thread=None, map_func=None, *, buffer_size=200, strict=False):
        """
        Args:
            ds (DataFlow): the dataflow to map
            num_thread (int): number of threads to use
            map_func (callable): datapoint -> datapoint | None. Return None to
                discard/skip the datapoint.
            buffer_size (int): number of datapoints in the buffer
            strict (bool): use "strict mode", see notes above.
        """
        if strict:
            # In strict mode, buffer size cannot be larger than the total number of datapoints
            try:
                buffer_size = min(buffer_size, len(ds))
            except Exception:  # ds may not have a length
                pass

        super(MultiThreadMapData, self).__init__(ds, buffer_size, strict)
        assert num_thread > 0, num_thread

        self._strict = strict
        self.num_thread = num_thread
        self.map_func = map_func
        self._threads = []
        self._evt = None

    def reset_state(self):
        super(MultiThreadMapData, self).reset_state()
        if self._threads:
            self._threads[0].stop()
            for t in self._threads:
                t.join()

        self._in_queue = queue.Queue()
        self._out_queue = queue.Queue()
        self._evt = threading.Event()
        self._threads = [MultiThreadMapData._Worker(
            self._in_queue, self._out_queue, self._evt, self.map_func)
            for _ in range(self.num_thread)]
        for t in self._threads:
            t.start()

        self._guard = DataFlowReentrantGuard()

        # Call once at the beginning, to ensure inq+outq has a total of buffer_size elements
        self._fill_buffer()

    def _recv(self):
        return self._out_queue.get()

    def _send(self, dp):
        self._in_queue.put(dp)

    def __iter__(self):
        with self._guard:
            yield from super(MultiThreadMapData, self).__iter__()

    def __del__(self):
        if self._evt is not None:
            self._evt.set()
        for p in self._threads:
            p.stop()
            p.join(timeout=5.0)
            # if p.is_alive():
            #     logger.warn("Cannot join thread {}.".format(p.name))


class MultiProcessMapDataZMQ(_ParallelMapData, _MultiProcessZMQDataFlow):
    """
    Same as :class:`MapData`, but start processes to run the mapping function,
    and communicate with ZeroMQ pipe.

    The semantics of this class is **identical** to :class:`MapData` except for the ordering.
    Processes run in parallel and can take different time to run the
    mapping function. Therefore the order of datapoints won't be preserved.

    When ``strict=True``, ``MultiProcessMapData(df, ...)``
    is guaranteed to produce the exact set of data as ``MapData(df, ...)``,
    if both are iterated until ``StopIteration``. But the produced data will have different ordering.
    The behavior of strict mode is undefined if the given dataflow ``df`` is infinite.

    When ``strict=False``, the data that's produced by ``MultiProcessMapData(df, ...)``
    is a reordering of the data produced by ``RepeatedData(MapData(df, ...), -1)``.
    In other words, first pass of ``MultiProcessMapData.__iter__`` may contain
    datapoints from the second pass of ``df.__iter__``.
    """
    class _Worker(mp.Process):
        def __init__(self, identity, map_func, pipename, hwm):
            super(MultiProcessMapDataZMQ._Worker, self).__init__()
            self.identity = identity
            self.map_func = map_func
            self.pipename = pipename
            self.hwm = hwm

        def run(self):
            enable_death_signal(_warn=self.identity == b'0')
            ctx = zmq.Context()
            socket = ctx.socket(zmq.REP)
            socket.setsockopt(zmq.IDENTITY, self.identity)
            socket.set_hwm(self.hwm)
            socket.connect(self.pipename)

            while True:
                dp = loads(socket.recv(copy=False))
                dp = self.map_func(dp)
                socket.send(dumps(dp), copy=False)

    def __init__(self, ds, num_proc=None, map_func=None, *, buffer_size=200, strict=False):
        """
        Args:
            ds (DataFlow): the dataflow to map
            num_proc(int): number of threads to use
            map_func (callable): datapoint -> datapoint | None. Return None to
                discard/skip the datapoint.
            buffer_size (int): number of datapoints in the buffer
            strict (bool): use "strict mode", see notes above.
        """
        if strict:
            # In strict mode, buffer size cannot be larger than the total number of datapoints
            try:
                buffer_size = min(buffer_size, len(ds))
            except Exception:  # ds may not have a length
                pass

        _ParallelMapData.__init__(self, ds, buffer_size, strict)
        _MultiProcessZMQDataFlow.__init__(self)
        assert num_proc > 0, num_proc
        self.num_proc = num_proc
        self.map_func = map_func
        self._strict = strict
        self._procs = []

    def _create_worker(self, id, pipename, hwm):
        return MultiProcessMapDataZMQ._Worker(id, self.map_func, pipename, hwm)

    def reset_state(self):
        _MultiProcessZMQDataFlow.reset_state(self)
        _ParallelMapData.reset_state(self)
        self._guard = DataFlowReentrantGuard()

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.set_hwm(self._buffer_size * 2)
        pipename = _get_pipe_name('dataflow-map')
        _bind_guard(self.socket, pipename)

        self._proc_ids = [u'{}'.format(k).encode('utf-8') for k in range(self.num_proc)]
        worker_hwm = int(self._buffer_size * 2 // self.num_proc)
        self._procs = [self._create_worker(self._proc_ids[k], pipename, worker_hwm)
                       for k in range(self.num_proc)]

        self._start_processes()
        self._fill_buffer()     # pre-fill the bufer

    def _send(self, dp):
        msg = [b"", dumps(dp)]
        self.socket.send_multipart(msg, copy=False)

    def _recv(self):
        msg = self.socket.recv_multipart(copy=False)
        dp = loads(msg[1])
        return dp

    def __iter__(self):
        with self._guard, _zmq_catch_error(type(self).__name__):
            yield from super(MultiProcessMapDataZMQ, self).__iter__()


class MultiProcessMapAndBatchDataZMQ(_MultiProcessZMQDataFlow):
    """
    Similar to :class:`MultiProcessMapDataZMQ`, except that this DataFlow
    also does batching in parallel in the worker processes.
    Therefore it can be helpful if you wish to hide the latency of batching.

    When `nr_proc==1`, the behavior of this class is identical to
    `BatchData(MapData(ds, map_func), batch_size)`.

    When `nr_proc>1`, the datapoints may be grouped in arbitrary order,
    or grouped with datapoints from a different pass of the given dataflow.
    """

    class _Dispatcher(mp.Process):
        def __init__(self, ds, pipename, hwm):
            super(MultiProcessMapAndBatchDataZMQ._Dispatcher, self).__init__()
            self.ds = RepeatedData(ds, -1)
            self.pipename = pipename
            self.hwm = hwm

        def run(self):
            enable_death_signal()
            ctx = zmq.Context()
            socket = ctx.socket(zmq.PUSH)
            socket.set_hwm(self.hwm)
            socket.bind(self.pipename)
            self.ds.reset_state()
            for dp in self.ds:
                socket.send(dumps(dp), copy=False)

    class _Worker(mp.Process):
        def __init__(self, identity, map_func, input_pipe, result_pipe, hwm, batch_size):
            super(MultiProcessMapAndBatchDataZMQ._Worker, self).__init__()
            self.identity = identity
            self.map_func = map_func
            self.input_pipe = input_pipe
            self.result_pipe = result_pipe
            self.hwm = hwm
            self.batch_size = batch_size

        def run(self):
            enable_death_signal(_warn=self.identity == b'0')
            ctx = zmq.Context()

            # recv jobs
            socket = ctx.socket(zmq.PULL)
            socket.setsockopt(zmq.IDENTITY, self.identity)
            socket.set_hwm(self.hwm * self.batch_size)
            socket.connect(self.input_pipe)

            # send results
            out_socket = ctx.socket(zmq.PUSH)
            out_socket.set_hwm(max(self.hwm, 5))
            out_socket.connect(self.result_pipe)

            batch = []
            while True:
                dp = loads(socket.recv(copy=False))
                dp = self.map_func(dp)
                if dp is not None:
                    batch.append(dp)
                    if len(batch) == self.batch_size:
                        dp = BatchData.aggregate_batch(batch)
                        out_socket.send(dumps(dp), copy=False)
                        del batch[:]

    def __init__(self, ds, num_proc, map_func, batch_size, buffer_size=None):
        """
        Args:
            ds (DataFlow): the dataflow to map
            num_proc(int): number of threads to use
            map_func (callable): datapoint -> datapoint | None. Return None to
                discard/skip the datapoint.
            batch_size (int): batch size
            buffer_size (int): number of datapoints (not batched) in the buffer.
                Defaults to batch_size * 10
        """
        super(MultiProcessMapAndBatchDataZMQ, self).__init__()
        assert batch_size < buffer_size
        self.ds = ds
        self.num_proc = num_proc
        self.map_func = map_func
        self.batch_size = batch_size
        if buffer_size is None:
            buffer_size = batch_size * 10
        self.buffer_size = buffer_size

    def reset_state(self):
        _MultiProcessZMQDataFlow.reset_state(self)
        self._guard = DataFlowReentrantGuard()

        job_pipe = _get_pipe_name("dataflow_MaB_job")
        result_pipe = _get_pipe_name("dataflow_MaB_result")

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PULL)
        self.socket.set_hwm(max(5, self.buffer_size // self.batch_size))
        _bind_guard(self.socket, result_pipe)

        dispatcher = MultiProcessMapAndBatchDataZMQ._Dispatcher(self.ds, job_pipe, self.buffer_size)

        self._proc_ids = [u'{}'.format(k).encode('utf-8') for k in range(self.num_proc)]
        worker_hwm = max(3, self.buffer_size // self.num_proc // self.batch_size)
        self._procs = [MultiProcessMapAndBatchDataZMQ._Worker(
            self._proc_ids[k], self.map_func, job_pipe, result_pipe, worker_hwm, self.batch_size)
            for k in range(self.num_proc)]

        self._procs.append(dispatcher)
        self._start_processes()

    def __iter__(self):
        with self._guard, _zmq_catch_error(type(self).__name__):
            while True:
                yield loads(self.socket.recv(copy=False))


def _pool_map(data):
    global SHARED_ARR, WORKER_ID, MAP_FUNC
    res = MAP_FUNC(data)
    if res is None:
        return None
    shared = np.reshape(SHARED_ARR, res.shape)
    assert shared.dtype == res.dtype
    shared[:] = res
    return WORKER_ID


# TODO shutdown pool, improve speed.
class MultiProcessMapDataComponentSharedArray(DataFlow):
    """
    Similar to :class:`MapDataComponent`, but perform IPC by shared memory,
    therefore more efficient when data (result of map_func) is large.
    It requires `map_func` to always return a numpy array of fixed shape and dtype, or None.
    """
    def __init__(self, ds, nr_proc, map_func, output_shape, output_dtype, index=0):
        """
        Args:
            ds (DataFlow): the dataflow to map on
            nr_proc(int): number of processes
            map_func (data component -> ndarray | None): the mapping function
            output_shape (tuple): the shape of the output of map_func
            output_dtype (np.dtype): the type of the output of map_func
            index (int): the index of the datapoint component to map on.
        """
        self.ds = ds
        self.nr_proc = nr_proc
        self.map_func = map_func
        self.output_shape = output_shape
        self.output_dtype = np.dtype(output_dtype).type
        self.index = index

        self._shared_mem = [self._create_shared_arr() for k in range(nr_proc)]
        id_queue = mp.Queue()
        for k in range(nr_proc):
            id_queue.put(k)

        def _init_pool(arrs, queue, map_func):
            id = queue.get()
            global SHARED_ARR, WORKER_ID, MAP_FUNC
            SHARED_ARR = arrs[id]
            WORKER_ID = id
            MAP_FUNC = map_func

        self._pool = mp.pool.Pool(
            processes=nr_proc,
            initializer=_init_pool,
            initargs=(self._shared_mem, id_queue, map_func))

    def _create_shared_arr(self):
        TYPE = {
            np.float32: ctypes.c_float,
            np.float64: ctypes.c_double,
            np.uint8: ctypes.c_uint8,
            np.int8: ctypes.c_int8,
            np.int32: ctypes.c_int32,
        }
        ctype = TYPE[self.output_dtype]
        arr = mp.RawArray(ctype, int(np.prod(self.output_shape)))
        return arr

    def __len__(self):
        return len(self.ds)

    def reset_state(self):
        self.ds.reset_state()
        self._guard = DataFlowReentrantGuard()

    def __iter__(self):
        ds_itr = _repeat_iter(self.ds.get_data)
        with self._guard:
            while True:
                dps = []
                for k in range(self.nr_proc):
                    dps.append(copy.copy(next(ds_itr)))
                to_map = [x[self.index] for x in dps]
                res = self._pool.map_async(_pool_map, to_map)

                for index in res.get():
                    if index is None:
                        continue
                    arr = np.reshape(self._shared_mem[index], self.output_shape)
                    dp = dps[index]
                    dp[self.index] = arr.copy()
                    yield dp


# alias
MultiProcessMapData = MultiProcessMapDataZMQ
MultiProcessMapAndBatchData = MultiProcessMapAndBatchDataZMQ


if __name__ == '__main__':
    import time

    class Zero(DataFlow):
        def __init__(self, size):
            self._size = size

        def __iter__(self):
            for k in range(self._size):
                yield [k]

        def __len__(self):
            return self._size

    def f(x):
        if x[0] < 10:
            time.sleep(1)
        return x

    ds = Zero(100)
    ds = MultiThreadMapData(ds, 50, f, buffer_size=50, strict=True)
    ds.reset_state()
    for idx, k in enumerate(ds):
        print("Bang!", k)
        if idx == 100:
            break
    print("END!")



# -*- coding: utf-8 -*-
# File: cifar.py

#         Yukun Chen <cykustc@gmail.com>

import numpy as np
import os
import pickle
import tarfile

from ...utils import logger
from ...utils.fs import download, get_dataset_path
from ..base import RNGDataFlow

__all__ = ['CifarBase', 'Cifar10', 'Cifar100']


DATA_URL_CIFAR_10 = ('http://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz', 170498071)
DATA_URL_CIFAR_100 = ('http://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz', 169001437)


def maybe_download_and_extract(dest_directory, cifar_classnum):
    """Download and extract the tarball from Alex's website. Copied from tensorflow example """
    assert cifar_classnum == 10 or cifar_classnum == 100
    if cifar_classnum == 10:
        cifar_foldername = 'cifar-10-batches-py'
    else:
        cifar_foldername = 'cifar-100-python'
    if os.path.isdir(os.path.join(dest_directory, cifar_foldername)):
        logger.info("Found cifar{} data in {}.".format(cifar_classnum, dest_directory))
        return
    else:
        DATA_URL = DATA_URL_CIFAR_10 if cifar_classnum == 10 else DATA_URL_CIFAR_100
        filename = DATA_URL[0].split('/')[-1]
        filepath = os.path.join(dest_directory, filename)
        download(DATA_URL[0], dest_directory, expect_size=DATA_URL[1])
        tarfile.open(filepath, 'r:gz').extractall(dest_directory)


def read_cifar(filenames, cifar_classnum):
    assert cifar_classnum == 10 or cifar_classnum == 100
    ret = []
    for fname in filenames:
        fo = open(fname, 'rb')
        dic = pickle.load(fo, encoding='bytes')
        data = dic[b'data']
        if cifar_classnum == 10:
            label = dic[b'labels']
            IMG_NUM = 10000  # cifar10 data are split into blocks of 10000
        else:
            label = dic[b'fine_labels']
            IMG_NUM = 50000 if 'train' in fname else 10000
        fo.close()
        for k in range(IMG_NUM):
            img = data[k].reshape(3, 32, 32)
            img = np.transpose(img, [1, 2, 0])
            ret.append([img, label[k]])
    return ret


def get_filenames(dir, cifar_classnum):
    assert cifar_classnum == 10 or cifar_classnum == 100
    if cifar_classnum == 10:
        train_files = [os.path.join(
            dir, 'cifar-10-batches-py', 'data_batch_%d' % i) for i in range(1, 6)]
        test_files = [os.path.join(
            dir, 'cifar-10-batches-py', 'test_batch')]
        meta_file = os.path.join(dir, 'cifar-10-batches-py', 'batches.meta')
    elif cifar_classnum == 100:
        train_files = [os.path.join(dir, 'cifar-100-python', 'train')]
        test_files = [os.path.join(dir, 'cifar-100-python', 'test')]
        meta_file = os.path.join(dir, 'cifar-100-python', 'meta')
    return train_files, test_files, meta_file


def _parse_meta(filename, cifar_classnum):
    with open(filename, 'rb') as f:
        obj = pickle.load(f)
        return obj['label_names' if cifar_classnum == 10 else 'fine_label_names']


class CifarBase(RNGDataFlow):
    """
    Produces [image, label] in Cifar10/100 dataset,
    image is 32x32x3 in the range [0,255].
    label is an int.
    """
    def __init__(self, train_or_test, shuffle=None, dir=None, cifar_classnum=10):
        """
        Args:
            train_or_test (str): 'train' or 'test'
            shuffle (bool): defaults to True for training set.
            dir (str): path to the dataset directory
            cifar_classnum (int): 10 or 100
        """
        assert train_or_test in ['train', 'test']
        assert cifar_classnum == 10 or cifar_classnum == 100
        self.cifar_classnum = cifar_classnum
        if dir is None:
            dir = get_dataset_path('cifar{}_data'.format(cifar_classnum))
        maybe_download_and_extract(dir, self.cifar_classnum)
        train_files, test_files, meta_file = get_filenames(dir, cifar_classnum)
        if train_or_test == 'train':
            self.fs = train_files
        else:
            self.fs = test_files
        for f in self.fs:
            if not os.path.isfile(f):
                raise ValueError('Failed to find file: ' + f)
        self._label_names = _parse_meta(meta_file, cifar_classnum)
        self.train_or_test = train_or_test
        self.data = read_cifar(self.fs, cifar_classnum)
        self.dir = dir

        if shuffle is None:
            shuffle = train_or_test == 'train'
        self.shuffle = shuffle

    def __len__(self):
        return 50000 if self.train_or_test == 'train' else 10000

    def __iter__(self):
        idxs = np.arange(len(self.data))
        if self.shuffle:
            self.rng.shuffle(idxs)
        for k in idxs:
            # since cifar is quite small, just do it for safety
            yield self.data[k]

    def get_per_pixel_mean(self, names=('train', 'test')):
        """
        Args:
            names (tuple[str]): the names ('train' or 'test') of the datasets

        Returns:
            a mean image of all images in the given datasets, with size 32x32x3
        """
        for name in names:
            assert name in ['train', 'test'], name
        train_files, test_files, _ = get_filenames(self.dir, self.cifar_classnum)
        all_files = []
        if 'train' in names:
            all_files.extend(train_files)
        if 'test' in names:
            all_files.extend(test_files)
        all_imgs = [x[0] for x in read_cifar(all_files, self.cifar_classnum)]
        arr = np.array(all_imgs, dtype='float32')
        mean = np.mean(arr, axis=0)
        return mean

    def get_label_names(self):
        """
        Returns:
            [str]: name of each class.
        """
        return self._label_names

    def get_per_channel_mean(self, names=('train', 'test')):
        """
        Args:
            names (tuple[str]): the names ('train' or 'test') of the datasets

        Returns:
            An array of three values as mean of each channel, for all images in the given datasets.
        """
        mean = self.get_per_pixel_mean(names)
        return np.mean(mean, axis=(0, 1))


class Cifar10(CifarBase):
    """
    Produces [image, label] in Cifar10 dataset,
    image is 32x32x3 in the range [0,255].
    label is an int.
    """
    def __init__(self, train_or_test, shuffle=None, dir=None):
        """
        Args:
            train_or_test (str): either 'train' or 'test'.
            shuffle (bool): shuffle the dataset, default to shuffle in training
        """
        super(Cifar10, self).__init__(train_or_test, shuffle, dir, 10)


class Cifar100(CifarBase):
    """ Similar to Cifar10"""
    def __init__(self, train_or_test, shuffle=None, dir=None):
        super(Cifar100, self).__init__(train_or_test, shuffle, dir, 100)


if __name__ == '__main__':
    ds = Cifar10('train')
    mean = ds.get_per_channel_mean()
    print(mean)

    import cv2
    ds.reset_state()
    for i, dp in enumerate(ds):
        if i == 100:
            break
        img = dp[0]
        cv2.imwrite("{:04d}.jpg".format(i), img)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tqdm import tqdm
from datasets.arrow_dataset import Dataset as ArrowDataset


def convert_glue_samples_to_dataset(samples, is_test=False):
    idx_list = []
    text_a_list = []
    text_b_list = []
    label_list = []

    for guid, text_a, text_b, label in tqdm(samples):
        idx_list.append(guid)
        text_a_list.append(text_a)
        text_b_list.append(text_b)
        if not is_test:
            label_list.append(label)

    data_dict = {
        'idx': idx_list,
        'sentence1': text_a_list,
    }
    if any(text_b_list):
        data_dict['sentence2'] = text_b_list
    if not is_test:
        data_dict['label'] = label_list

    dataset = ArrowDataset.from_dict(data_dict)

    return dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
from collections import OrderedDict
from typing import Type

import dill
import numpy as np
import torch
import torch.functional as F
import torch.nn as nn
from loguru import logger
from sklearn.metrics import f1_score
from torch.nn import CrossEntropyLoss, MSELoss
from tqdm import tqdm

from theta.nlp.arguments import (DataArguments, ModelArguments, TaskArguments,
                                 TrainingArguments)
from theta.nlp.data.samples import GlueSamples
from transformers import AutoModelForSequenceClassification

from .task import BaseDataset, BaseTask, TaskData, TaskRunner, TransformerModel


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def simple_accuracy(preds, labels):
    return (preds == labels).mean()


#  def acc_and_f1(preds, labels, average="micro"):
#      correct = np.sum((preds == labels).astype(int))
#      acc = correct / preds.shape[0]
#      f1 = f1_score(y_true=labels,
#                    y_pred=preds,
#                    average=average,
#                    zero_division=1)
#      acc_and_f1 = (acc + f1) / 2
#      return {"acc": acc, "f1": f1, "acc_and_f1": acc_and_f1}


def p_r_f1(preds, labels):
    if len(preds.shape) > 1 and len(labels.shape) > 1:
        tp = np.sum(((preds + labels) == 2).astype(int))
        fp = np.sum(((preds - labels) == 1).astype(int))
        fn = np.sum(((labels - preds) == 1).astype(int))
    else:
        tp = np.sum((preds == labels).astype(int))
        num_true = labels.shape[0]
        fn = num_true - tp
        num_pred = preds.shape[0]
        fp = num_pred - tp

    p = tp / (tp + fp) if tp + fp != 0 else 0.0
    r = tp / (tp + fn) if tp + fn != 0 else 0.0
    f1 = 2 * p * r / (p + r) if (p + r) != 0 else 0.0

    #  return {"acc": acc, "f1": f1, "acc_and_f1": acc_and_f1}
    return (p, r, f1)


# ------------------------------ Dataset ------------------------------
class GlueDataset(BaseDataset):
    """
    完成模型输入数据的编码工作
    """
    def __init__(self, *args, **kwargs):
        super(GlueDataset, self).__init__(*args, **kwargs)

    def _encode_item(self, x):
        guid, text_a, text_b, labels = x

        # -------- input_ids, attention_mask, token_type_ids --------
        text_pair = [(text_a, text_b)] if text_b is not None else [text_a]
        encodings = self.tokenizer.batch_encode_plus(
            text_pair,
            padding=self.data_args.padding,
            max_length=self.data_args.max_length,
            add_special_tokens=True,
            truncation=True,
            return_offsets_mapping=True)
        input_ids = torch.from_numpy(
            np.array(encodings.input_ids, dtype=np.int64))[0]
        attention_mask = torch.from_numpy(
            np.array(encodings.attention_mask, dtype=np.int64))[0]
        token_type_ids = torch.from_numpy(
            np.array(encodings.token_type_ids, dtype=np.int64))[0]

        # -------- labels --------
        if labels is not None:
            if isinstance(labels, list):
                encoded_labels = [0] * len(self.label2id)
                for x in labels:
                    encoded_labels[self.label2id[x]] = 1
                labels = torch.from_numpy(
                    np.array(encoded_labels, dtype=np.float32))
            else:
                if labels:
                    encoded_labels = self.label2id[labels]
                else:
                    encoded_labels = 0
                labels = torch.from_numpy(
                    np.array(encoded_labels, dtype=np.int64))

        return {
            'input_ids': input_ids,
            'attention_mask': attention_mask,
            'token_type_ids': token_type_ids,
            'labels': labels
        }

    @classmethod
    def collate_fn(cls, batch):
        stacked_batch = {}
        #  for key in ['input_ids', 'attention_mask', 'token_type_ids', 'labels']:
        #      key_batch = [e[key] for e in batch if e[key] is not None]
        #      if key_batch:
        #          batch_values = torch.stack(key_batch)
        #          stacked_batch[key] = batch_values
        #      else:
        #          stacked_batch[key] = None
        #  return stacked_batch

        not_none_tensor_keys = [
            'input_ids', 'attention_mask', 'token_type_ids'
        ]
        maybe_none_tensor_keys = ['labels']
        not_tensor_keys = []

        # not None tensors
        for key in not_none_tensor_keys:
            key_batch = [e[key] for e in batch if e[key] is not None]
            #  logger.info(f"key: {key} key_batch: {key_batch}")
            batch_values = torch.stack(key_batch)
            stacked_batch[key] = batch_values
        # maybe None tensors
        for key in maybe_none_tensor_keys:
            key_batch = [e[key] for e in batch if e[key] is not None]
            if key_batch:
                batch_values = torch.stack(key_batch)
                stacked_batch[key] = batch_values
            else:
                stacked_batch[key] = None
        # not tensors
        for key in not_tensor_keys:
            key_batch = [e[key] for e in batch if e[key] is not None]
            stacked_batch[key] = key_batch

        return stacked_batch


# ------------------------------ TaskData ------------------------------
class GlueData(TaskData):
    """
    指定任务专属的Dataset类型，并提供训练、验证、测试集实例。
    """
    def __init__(self, *args, **kwargs):
        super(GlueData, self).__init__(*args, **kwargs)

    def build_train_dataset(self):
        return GlueDataset(self.data_args, self._splitted_train_samples.rows,
                           self.label2id, self.tokenizer)

    def build_val_dataset(self):
        return GlueDataset(self.data_args, self._splitted_val_samples.rows,
                           self.label2id, self.tokenizer)

    def build_test_dataset(self):
        return GlueDataset(self.data_args, self.test_samples.rows,
                           self.label2id, self.tokenizer)


class MyGlueModel(TransformerModel):
    def __init__(
        self,
        model_name_or_path,
        num_labels,
        tokenizer=None,
        dropout_prob=0.1,
        #  loss_type='CrossEntropyLoss',
        loss_type='FocalLoss',
        #  loss_type='LabelSmoothingCrossEntropy',
        #  loss_type='DiceLoss',
        **kwargs):

        # for TransformerModel.load_from_config()
        self.num_labels = num_labels
        super(MyGlueModel,
              self).__init__(model_name_or_path,
                             tokenizer=tokenizer,
                             automodel_cls=AutoModelForSequenceClassification)
        #  config = self.config
        #  if self._is_xlnet():
        #      from transformers.modeling_utils import SequenceSummary
        #      self.sequence_summary = SequenceSummary(config)
        #      self.logits_proj = nn.Linear(config.d_model, self.num_labels)
        #  else:
        #      self.dropout = nn.Dropout(config.hidden_dropout_prob)
        #      self.classifier = nn.Linear(config.hidden_size, self.num_labels)

        self.init_weights()

    def load_from_pretrained(self, model_path):
        #  automodel_cls = AutoModelForSequenceClassification
        #  super(MyGlueModel,
        #        self)._load_from_pretrained(model_path,
        #                                    automodel_cls=automodel_cls,
        #                                    num_labels=self.num_labels)
        setattr(self.config, 'num_labels', self.num_labels)
        self.transformer = AutoModelForSequenceClassification.from_pretrained(
            model_path, config=self.config)
        #  self.transformer = AutoModelForSequenceClassification.from_pretrained(
        #      model_path, num_labels=self.num_labels)

    def load_from_config(self):
        setattr(self.config, 'num_labels', self.num_labels)
        self.transformer = AutoModelForSequenceClassification.from_config(
            self.config)

    def forward(self,
                input_ids=None,
                attention_mask=None,
                token_type_ids=None,
                labels=None):

        outputs = self.transformer(input_ids,
                                   attention_mask=attention_mask,
                                   token_type_ids=token_type_ids,
                                   labels=labels,
                                   return_dict=True)
        #  logger.warning(f"outputs[0].shape: {outputs[0].shape}")
        #  logger.warning(f"outputs: {outputs}")
        # logger.warning(f"outputs[1]: {outputs[1]}")

        #  if labels is not None:
        #      loss = outputs.loss
        #      logits = outputs.logits
        #      return (loss, logits)
        #  else:
        #      logits = outputs.logits
        #      return logits

        #          if self._is_xlnet():
        #              output = outputs[0]
        #              output = self.sequence_summary(output)
        #              logits = self.logits_proj(output)
        #          #  elif self._is_electra():
        #          #      # ElectraConfig last_hidden_state
        #          #      pooled_output = outputs[0]
        #          #
        #          #      pooled_output = self.dropout(pooled_output)
        #          #      logits = self.classifier(pooled_output)
        #          else:
        #              # BERT
        #              pooled_output = outputs.pooler_output
        #
        #              pooled_output = self.dropout(pooled_output)
        #              logits = self.classifier(pooled_output)
        #  #
        #          if labels is not None:
        #              loss_fct = CrossEntropyLoss()
        #              loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
        #              return (loss, logits)
        #          else:
        #              return logits

        loss = outputs.loss
        logits = outputs.logits
        if labels is not None:
            return (loss, logits)
        else:
            return logits


# ------------------------------ TaskRunner ------------------------------
class GlueRunner(TaskRunner):
    """
    任务专属模型定义
    """
    def __init__(self, task_args, glue_labels):
        super(GlueRunner, self).__init__(**task_args.to_dict())
        logger.warning(f"glue_labels: {glue_labels}")
        self.glue_labels = glue_labels
        self.num_labels = len(self.glue_labels)
        self.label2id = {x: i for i, x in enumerate(glue_labels)}
        self.id2label = {i: x for i, x in enumerate(glue_labels)}
        self.type_weights = np.ones(self.num_labels)

        model_args = task_args.model_args

        logger.warning(f"num_labels: {self.num_labels}")
        self.model = MyGlueModel(
            model_name_or_path=model_args.model_name_or_path
            if model_args.model_name_or_path else model_args.checkpoint_path,
            num_labels=self.num_labels)
        #  config = self.transformer.config
        #  if self._is_xlnet():
        #      from transformers.modeling_utils import SequenceSummary
        #      self.sequence_summary = SequenceSummary(config)
        #      self.logits_proj = nn.Linear(config.d_model, num_labels)
        #  else:
        #      self.dropout = nn.Dropout(config.hidden_dropout_prob)
        #      self.classifier = nn.Linear(config.hidden_size, num_labels)

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    #  def forward(self, batch, batch_idx):
    #  def forward(self,
    #              input_ids=None,
    #              attention_mask=None,
    #              token_type_ids=None,
    #              labels=None):
    #
    #      return self.model(*args, **kwargs)
#          outputs = self.transformer(input_ids,
#                                           attention_mask=attention_mask,
#                                           token_type_ids=token_type_ids,
#                                           return_dict=True)
#          #  logger.warning(f"outputs[0].shape: {outputs[0].shape}")
#          #  logger.warning(f"outputs: {outputs}")
#          # logger.warning(f"outputs[1]: {outputs[1]}")
#
#          #  if labels is not None:
#          #      loss = outputs.loss
#          #      logits = outputs.logits
#          #      return (loss, logits)
#          #  else:
#          #      logits = outputs.logits
#          #      return logits
#
#          if self._is_xlnet():
#              output = outputs[0]
#              output = self.sequence_summary(output)
#              logits = self.logits_proj(output)
#          elif self._is_electra():
#              # ElectraConfig last_hidden_state
#              pooled_output = outputs[0]
#
#              pooled_output = self.dropout(pooled_output)
#              logits = self.classifier(pooled_output)
#          else:
#              # BERT
#              pooled_output = outputs.pooler_output
#
#              pooled_output = self.dropout(pooled_output)
#              logits = self.classifier(pooled_output)
#  #
#          if labels is not None:
#              loss_fct = CrossEntropyLoss()
#              loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
#              return (loss, logits)
#          else:
#              return logits

    def training_step(self, batch, batch_idx):
        outputs = self.forward(**batch)
        loss = outputs[0]

        self.log('train_loss', loss, on_step=True)
        #  self.log('lr', self.hparams.lr, on_step=True)

        return OrderedDict({
            "loss": loss,
        })

    def validation_step(self, batch, batch_idx):
        #  logger.warning(f"batch: {batch}")
        #  logger.info(f"batch['input_ids'].shape: {batch['input_ids'].shape}")
        #  logger.info(
        #      f"batch['attention_mask'].shape: {batch['attention_mask'].shape}")
        #  logger.info(
        #      f"batch['token_type_ids'].shape: {batch['token_type_ids'].shape}")
        #  logger.info(f"batch['labels'].shape: {batch['labels'].shape}")
        val_loss, logits = self.forward(**batch)

        #  preds = np.argmax(logits, axis=1)

        batch_labels = batch['labels']
        batch_labels = batch_labels.cpu().numpy().astype(int)

        is_multi_label_classification = False
        if len(batch_labels.shape) == 1:
            is_multi_label_classification = False
        else:
            is_multi_label_classification = True
        #  logger.info(
        #      f"is_multi_label_classification: {is_multi_label_classification}")

        if is_multi_label_classification:
            confidence = 0.5
            batch_probs = torch.sigmoid(logits).cpu().numpy()
            batch_preds = (batch_probs > confidence).astype(int)
            val_acc, val_recall, val_f1 = p_r_f1(batch_preds, batch_labels)

        else:
            #  correct_count = torch.sum(batch_labels == batch_preds).float()
            #  if self.on_gpu:
            #      correct_cout = correct_count.cuda(val_loss.device.index)
            batch_probs = torch.softmax(logits, -1)
            batch_preds = torch.argmax(logits, dim=1).cpu().numpy()
            val_acc, val_recall, val_f1 = p_r_f1(batch_preds, batch_labels)

        self.log('val_loss', val_loss, on_step=True)
        self.log('val_acc', val_acc, on_step=True)
        self.log('val_recall', val_recall, on_step=True)
        self.log('val_f1', val_f1, on_step=True)

        #  logger.info(f"val_loss: {val_loss:.4f}")

        return OrderedDict({
            'val_loss': val_loss,
            'val_acc': val_acc,
            'val_recall': val_recall,
            'val_f1': val_f1,
            'batch_probs': batch_probs,
            #  'correct_count': correct_count,
            'batch_size': batch_labels.shape[0]
        })

    def validation_epoch_end(self, outputs):
        #  val_acc = sum([out["correct_count"]
        #                 for out in outputs]).float() / sum(out["batch_size"]
        #                                                    for out in outputs)
        #  val_acc = sum([out["correct_count"]
        #                 for out in outputs]) / sum(out["batch_size"]
        #                                            for out in outputs)
        val_loss = sum([out["val_loss"] for out in outputs]) / len(outputs)
        val_acc = sum([out["val_acc"] for out in outputs]) / len(outputs)
        val_recall = sum([out["val_recall"] for out in outputs]) / len(outputs)
        val_f1 = sum([out["val_f1"] for out in outputs]) / len(outputs)
        logger.info(
            f"val_loss: {val_loss:.4f}, val_acc: {val_acc:.4f}, val_recall: {val_recall:.4f}, val_f1: {val_f1:.4f}"
        )

        self.log('val_loss', val_loss, on_epoch=True)
        self.log('val_acc', val_acc, on_epoch=True)
        self.log('val_recall', val_acc, on_epoch=True)
        self.log('val_f1', val_f1, on_epoch=True)

        eval_outputs = {
            'val_loss': val_loss,
            'val_acc': val_acc,
            'val_recall': val_recall,
            'val_f1': val_f1
        }
        self.save_best_model(eval_outputs)

    def test_step(self, batch, batch_idx):
        logits = self.forward(**batch)

        preds = torch.argmax(logits, dim=1)

        return OrderedDict({'preds': preds, 'logits': logits})

    def test_epoch_end(self, outputs):

        final_preds = torch.cat([out['preds'] for out in outputs])
        logger.info(f"preds: {final_preds.shape}")
        final_preds = final_preds.detach().cpu().numpy().tolist()

        final_logits = torch.cat([out['logits'] for out in outputs])
        logger.info(f"logits: {final_logits.shape}")
        final_logits = final_logits.detach().cpu().numpy().tolist()

        self.test_results = {'preds': final_preds, 'logits': final_logits}


# ------------------------------ Task ------------------------------
class GlueTask(BaseTask):
    #  def __init__(self, *args, **kwargs):
    #      super(GlueTask, self).__init__(*args, **kwargs)
    def __init__(self, args: Type[TaskArguments], data: Type[TaskData],
                 glue_labels: list):
        runner = GlueRunner(args, glue_labels)
        super(GlueTask, self).__init__(args, data, runner)

    def execute(self, *args, **kwargs):
        model_args = self.model_args
        data_args = self.data_args
        training_args = self.training_args
        remaining_args = self.remaining_args

        return super(GlueTask, self).execute(*args, **kwargs)

    def generate_submission(self):
        logger.warning(f"GlueTask.generate_submission().")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import torch
from loguru import logger
import numpy as np
from collections import defaultdict
from .ner_decodes import crf_decode, span_decode, mrc_decode


def calculate_metric(gt, predict):
    """
    计算 tp fp fn
    """
    tp, fp, fn = 0, 0, 0
    for entity_predict in predict:
        flag = 0
        for entity_gt in gt:
            if entity_predict[0] == entity_gt[0] and entity_predict[
                    1] == entity_gt[1]:
                flag = 1
                tp += 1
                break
        if flag == 0:
            fp += 1

    fn = len(gt) - tp

    return np.array([tp, fp, fn])


def get_p_r_f(tp, fp, fn):
    p = tp / (tp + fp) if tp + fp != 0 else 0
    r = tp / (tp + fn) if tp + fn != 0 else 0
    f1 = 2 * p * r / (p + r) if p + r != 0 else 0
    return np.array([p, r, f1])


def crf_evaluation(model, dev_info, device, ent2id):
    dev_loader, (dev_callback_info, type_weight) = dev_info

    pred_tokens = []

    for tmp_pred in get_base_out(model, dev_loader, device):
        pred_tokens.extend(tmp_pred[0])

    assert len(pred_tokens) == len(dev_callback_info)

    id2ent = {ent2id[key]: key for key in ent2id.keys()}

    role_metric = np.zeros([13, 3])

    mirco_metrics = np.zeros(3)

    for tmp_tokens, tmp_callback in zip(pred_tokens, dev_callback_info):

        text, gt_entities = tmp_callback

        tmp_metric = np.zeros([13, 3])

        pred_entities = crf_decode(tmp_tokens, text, id2ent)

        for idx, _type in enumerate(ENTITY_TYPES):
            if _type not in pred_entities:
                pred_entities[_type] = []

            tmp_metric[idx] += calculate_metric(gt_entities[_type],
                                                pred_entities[_type])

        role_metric += tmp_metric

    for idx, _type in enumerate(ENTITY_TYPES):
        temp_metric = get_p_r_f(role_metric[idx][0], role_metric[idx][1],
                                role_metric[idx][2])

        mirco_metrics += temp_metric * type_weight[_type]

    metric_str = f'[MIRCO] precision: {mirco_metrics[0]:.4f}, ' \
                 f'recall: {mirco_metrics[1]:.4f}, f1: {mirco_metrics[2]:.4f}'

    return metric_str, mirco_metrics[2]


def span_evaluation(model, dev_info, device, ent2id):
    dev_loader, (dev_callback_info, type_weight) = dev_info

    start_logits, end_logits = None, None

    model.eval()

    for tmp_pred in get_base_out(model, dev_loader, device):
        tmp_start_logits = tmp_pred[0].cpu().numpy()
        tmp_end_logits = tmp_pred[1].cpu().numpy()

        if start_logits is None:
            start_logits = tmp_start_logits
            end_logits = tmp_end_logits
        else:
            start_logits = np.append(start_logits, tmp_start_logits, axis=0)
            end_logits = np.append(end_logits, tmp_end_logits, axis=0)

    assert len(start_logits) == len(end_logits) == len(dev_callback_info)

    role_metric = np.zeros([13, 3])

    mirco_metrics = np.zeros(3)

    id2ent = {ent2id[key]: key for key in ent2id.keys()}

    for tmp_start_logits, tmp_end_logits, tmp_callback \
            in zip(start_logits, end_logits, dev_callback_info):

        text, gt_entities = tmp_callback

        tmp_start_logits = tmp_start_logits[1:1 + len(text)]
        tmp_end_logits = tmp_end_logits[1:1 + len(text)]

        pred_entities = span_decode(tmp_start_logits, tmp_end_logits, text,
                                    id2ent)

        for idx, _type in enumerate(ENTITY_TYPES):
            if _type not in pred_entities:
                pred_entities[_type] = []

            role_metric[idx] += calculate_metric(gt_entities[_type],
                                                 pred_entities[_type])

    for idx, _type in enumerate(ENTITY_TYPES):
        temp_metric = get_p_r_f(role_metric[idx][0], role_metric[idx][1],
                                role_metric[idx][2])

        mirco_metrics += temp_metric * type_weight[_type]

    metric_str = f'[MIRCO] precision: {mirco_metrics[0]:.4f}, ' \
                 f'recall: {mirco_metrics[1]:.4f}, f1: {mirco_metrics[2]:.4f}'

    return metric_str, mirco_metrics[2]


def mrc_evaluation(model, dev_info, device):
    dev_loader, (dev_callback_info, type_weight) = dev_info

    start_logits, end_logits = None, None

    model.eval()

    for tmp_pred in get_base_out(model, dev_loader, device):
        tmp_start_logits = tmp_pred[0].cpu().numpy()
        tmp_end_logits = tmp_pred[1].cpu().numpy()

        if start_logits is None:
            start_logits = tmp_start_logits
            end_logits = tmp_end_logits
        else:
            start_logits = np.append(start_logits, tmp_start_logits, axis=0)
            end_logits = np.append(end_logits, tmp_end_logits, axis=0)

    assert len(start_logits) == len(end_logits) == len(dev_callback_info)

    role_metric = np.zeros([13, 3])

    mirco_metrics = np.zeros(3)

    id2ent = {x: i for i, x in enumerate(ENTITY_TYPES)}

    for tmp_start_logits, tmp_end_logits, tmp_callback \
            in zip(start_logits, end_logits, dev_callback_info):

        text, text_offset, ent_type, gt_entities = tmp_callback

        tmp_start_logits = tmp_start_logits[text_offset:text_offset +
                                            len(text)]
        tmp_end_logits = tmp_end_logits[text_offset:text_offset + len(text)]

        pred_entities = mrc_decode(tmp_start_logits, tmp_end_logits, text)

        role_metric[id2ent[ent_type]] += calculate_metric(
            gt_entities, pred_entities)

    for idx, _type in enumerate(ENTITY_TYPES):
        temp_metric = get_p_r_f(role_metric[idx][0], role_metric[idx][1],
                                role_metric[idx][2])

        mirco_metrics += temp_metric * type_weight[_type]

    metric_str = f'[MIRCO] precision: {mirco_metrics[0]:.4f}, ' \
                  f'recall: {mirco_metrics[1]:.4f}, f1: {mirco_metrics[2]:.4f}'

    return metric_str, mirco_metrics[2]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .trainer import SpoTrainer, load_model, load_pretrained_model, load_pretrained_tokenizer
from .dataset import encode_examples 
from .args import get_args



# -*- coding: utf-8 -*-
# File: fs.py


import errno
import os
import tqdm
from six.moves import urllib

from . import logger
from .utils import execute_only_once

__all__ = ['mkdir_p', 'download', 'recursive_walk', 'get_dataset_path', 'normpath']


def mkdir_p(dirname):
    """ Like "mkdir -p", make a dir recursively, but do nothing if the dir exists

    Args:
        dirname(str):
    """
    assert dirname is not None
    if dirname == '' or os.path.isdir(dirname):
        return
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise e


def download(url, dir, filename=None, expect_size=None):
    """
    Download URL to a directory.
    Will figure out the filename automatically from URL, if not given.
    """
    mkdir_p(dir)
    if filename is None:
        filename = url.split('/')[-1]
    fpath = os.path.join(dir, filename)

    if os.path.isfile(fpath):
        if expect_size is not None and os.stat(fpath).st_size == expect_size:
            logger.info("File {} exists! Skip download.".format(filename))
            return fpath
        else:
            logger.warn("File {} exists. Will overwrite with a new download!".format(filename))

    def hook(t):
        last_b = [0]

        def inner(b, bsize, tsize=None):
            if tsize is not None:
                t.total = tsize
            t.update((b - last_b[0]) * bsize)
            last_b[0] = b
        return inner
    try:
        with tqdm.tqdm(unit='B', unit_scale=True, miniters=1, desc=filename) as t:
            fpath, _ = urllib.request.urlretrieve(url, fpath, reporthook=hook(t))
        statinfo = os.stat(fpath)
        size = statinfo.st_size
    except IOError:
        logger.error("Failed to download {}".format(url))
        raise
    assert size > 0, "Downloaded an empty file from {}!".format(url)

    if expect_size is not None and size != expect_size:
        logger.error("File downloaded from {} does not match the expected size!".format(url))
        logger.error("You may have downloaded a broken file, or the upstream may have modified the file.")

    # TODO human-readable size
    logger.info('Succesfully downloaded ' + filename + ". " + str(size) + ' bytes.')
    return fpath


def recursive_walk(rootdir):
    """
    Yields:
        str: All files in rootdir, recursively.
    """
    for r, dirs, files in os.walk(rootdir):
        for f in files:
            yield os.path.join(r, f)


def get_dataset_path(*args):
    """
    Get the path to some dataset under ``$TENSORPACK_DATASET``.

    Args:
        args: strings to be joined to form path.

    Returns:
        str: path to the dataset.
    """
    d = os.environ.get('TENSORPACK_DATASET', None)
    if d is None:
        d = os.path.join(os.path.expanduser('~'), 'tensorpack_data')
        if execute_only_once():
            logger.warn("Env var $TENSORPACK_DATASET not set, using {} for datasets.".format(d))
        if not os.path.isdir(d):
            mkdir_p(d)
            logger.info("Created the directory {}.".format(d))
    assert os.path.isdir(d), d
    return os.path.join(d, *args)


def normpath(path):
    """
    Normalizes a path to a folder by taking into consideration remote storages like Cloud storaged
    referenced by '://' at the beginning of the path.

    Args:
        args: path to be normalized.

    Returns:
        str: normalized path.
    """
    return path if '://' in path else os.path.normpath(path)


if __name__ == '__main__':
    download('http://dl.caffe.berkeleyvision.org/caffe_ilsvrc12.tar.gz', '.')



# -*- coding: utf-8 -*-
# File: format.py


import numpy as np
import os
import six

from ..utils import logger
from ..utils.argtools import log_once
from ..utils.serialize import loads
from ..utils.develop import create_dummy_class  # noqa
from ..utils.loadcaffe import get_caffe_pb
from ..utils.timer import timed_operation
from ..utils.utils import get_tqdm
from .base import DataFlowReentrantGuard, RNGDataFlow
from .common import MapData

__all__ = ['HDF5Data', 'LMDBData', 'LMDBDataDecoder',
           'CaffeLMDB', 'SVMLightData']

"""
Adapters for different data format.
"""


class HDF5Data(RNGDataFlow):
    """
    Zip data from different paths in an HDF5 file.

    Warning:
        The current implementation will load all data into memory. (TODO)
    """
# TODO

    def __init__(self, filename, data_paths, shuffle=True):
        """
        Args:
            filename (str): h5 data file.
            data_paths (list): list of h5 paths to zipped.
                For example `['images', 'labels']`.
            shuffle (bool): shuffle all data.
        """
        self.f = h5py.File(filename, 'r')
        logger.info("Loading {} to memory...".format(filename))
        self.dps = [self.f[k].value for k in data_paths]
        lens = [len(k) for k in self.dps]
        assert all(k == lens[0] for k in lens)
        self._size = lens[0]
        self.shuffle = shuffle

    def __len__(self):
        return self._size

    def __iter__(self):
        idxs = list(range(self._size))
        if self.shuffle:
            self.rng.shuffle(idxs)
        for k in idxs:
            yield [dp[k] for dp in self.dps]


class LMDBData(RNGDataFlow):
    """
    Read a LMDB database and produce (k,v) raw bytes pairs.
    The raw bytes are usually not what you're interested in.
    You might want to use
    :class:`LMDBDataDecoder` or apply a
    mapper function after :class:`LMDBData`.
    """
    def __init__(self, lmdb_path, shuffle=True, keys=None):
        """
        Args:
            lmdb_path (str): a directory or a file.
            shuffle (bool): shuffle the keys or not.
            keys (list[str] or str): list of str as the keys, used only when shuffle is True.
                It can also be a format string e.g. ``{:0>8d}`` which will be
                formatted with the indices from 0 to *total_size - 1*.

                If not given, it will then look in the database for ``__keys__`` which
                :func:`LMDBSerializer.save` used to store the list of keys.
                If still not found, it will iterate over the database to find
                all the keys.
        """
        self._lmdb_path = lmdb_path
        self._shuffle = shuffle

        self._open_lmdb()
        self._size = self._txn.stat()['entries']
        self._set_keys(keys)
        logger.info("Found {} entries in {}".format(self._size, self._lmdb_path))

        # Clean them up after finding the list of keys, since we don't want to fork them
        self._close_lmdb()

    def _set_keys(self, keys=None):
        def find_keys(txn, size):
            logger.warn("Traversing the database to find keys is slow. Your should specify the keys.")
            keys = []
            with timed_operation("Loading LMDB keys ...", log_start=True), \
                    get_tqdm(total=size) as pbar:
                for k in self._txn.cursor():
                    assert k[0] != b'__keys__'
                    keys.append(k[0])
                    pbar.update()
            return keys

        self.keys = self._txn.get(b'__keys__')
        if self.keys is not None:
            self.keys = loads(self.keys)
            self._size -= 1     # delete this item

        if self._shuffle:   # keys are necessary when shuffle is True
            if keys is None:
                if self.keys is None:
                    self.keys = find_keys(self._txn, self._size)
            else:
                # check if key-format like '{:0>8d}' was given
                if isinstance(keys, six.string_types):
                    self.keys = map(lambda x: keys.format(x), list(np.arange(self._size)))
                else:
                    self.keys = keys

    def _open_lmdb(self):
        self._lmdb = lmdb.open(self._lmdb_path,
                               subdir=os.path.isdir(self._lmdb_path),
                               readonly=True, lock=False, readahead=True,
                               map_size=1099511627776 * 2, max_readers=100)
        self._txn = self._lmdb.begin()

    def _close_lmdb(self):
        self._lmdb.close()
        del self._lmdb
        del self._txn

    def reset_state(self):
        self._guard = DataFlowReentrantGuard()
        super(LMDBData, self).reset_state()
        self._open_lmdb()  # open the LMDB in the worker process

    def __len__(self):
        return self._size

    def __iter__(self):
        with self._guard:
            if not self._shuffle:
                c = self._txn.cursor()
                for k, v in c:
                    if k != b'__keys__':
                        yield [k, v]
            else:
                self.rng.shuffle(self.keys)
                for k in self.keys:
                    v = self._txn.get(k)
                    yield [k, v]


class LMDBDataDecoder(MapData):
    """ Read a LMDB database with a custom decoder and produce decoded outputs."""
    def __init__(self, lmdb_data, decoder):
        """
        Args:
            lmdb_data: a :class:`LMDBData` instance.
            decoder (k,v -> dp | None): a function taking k, v and returning a datapoint,
                or return None to discard.
        """
        def f(dp):
            return decoder(dp[0], dp[1])
        super(LMDBDataDecoder, self).__init__(lmdb_data, f)


def CaffeLMDB(lmdb_path, shuffle=True, keys=None):
    """
    Read a Caffe-format LMDB file where each value contains a ``caffe.Datum`` protobuf.
    Produces datapoints of the format: [HWC image, label].

    Note that Caffe LMDB format is not efficient: it stores serialized raw
    arrays rather than JPEG images.

    Args:
        lmdb_path, shuffle, keys: same as :class:`LMDBData`.

    Example:
        .. code-block:: python

            ds = CaffeLMDB("/tmp/validation", keys='{:0>8d}')
    """

    cpb = get_caffe_pb()
    lmdb_data = LMDBData(lmdb_path, shuffle, keys)

    def decoder(k, v):
        try:
            datum = cpb.Datum()
            datum.ParseFromString(v)
            img = np.fromstring(datum.data, dtype=np.uint8)
            img = img.reshape(datum.channels, datum.height, datum.width)
        except Exception:
            log_once("Cannot read key {}".format(k), 'warn')
            return None
        return [img.transpose(1, 2, 0), datum.label]
    logger.warn("Caffe LMDB format doesn't store jpeg-compressed images, \
        it's not recommended due to its inferior performance.")
    return LMDBDataDecoder(lmdb_data, decoder)


class SVMLightData(RNGDataFlow):
    """ Read X,y from an SVMlight file, and produce [X_i, y_i] pairs. """

    def __init__(self, filename, shuffle=True):
        """
        Args:
            filename (str): input file
            shuffle (bool): shuffle the data
        """
        import sklearn.datasets  # noqa
        self.X, self.y = sklearn.datasets.load_svmlight_file(filename)
        self.X = np.asarray(self.X.todense())
        self.shuffle = shuffle

    def __len__(self):
        return len(self.y)

    def __iter__(self):
        idxs = np.arange(self.__len__())
        if self.shuffle:
            self.rng.shuffle(idxs)
        for id in idxs:
            yield [self.X[id, :], self.y[id]]


try:
    import h5py
except ImportError:
    HDF5Data = create_dummy_class('HDF5Data', 'h5py')   # noqa

try:
    import lmdb
except ImportError:
    for klass in ['LMDBData', 'LMDBDataDecoder', 'CaffeLMDB']:
        globals()[klass] = create_dummy_class(klass, 'lmdb')
